"""TcEx Framework Module"""

# standard library
import json
from collections.abc import Generator
from pathlib import Path

# third-party
from pydantic import ValidationError
from requests.exceptions import ProxyError

# first-party
from tcex.api.tc.v3._gen._options_abc import OptionsABC
from tcex.api.tc.v3._gen.model import PropertyModel
from tcex.pleb.cached_property import cached_property
from tcex.util.render.render import Render
from tcex.util.util import Util

# The OptionsData class is used to download the options data for a given object type and
# correct and/or update the data to ensure it is usable for the code generation process.
# The "plan" is to have a single class that handles downloading the data and updating it.
# The contents* methods download, update, and load to the model. The content_models method
# is typically the entry point to this class.


class OptionsDataObject(OptionsABC):
    """Gen Model Download Options Class"""

    def __init__(self, api_url: str, object_type: str):
        """Initialize instance properties."""
        self.api_url = api_url
        self.object_type = object_type

        # properties
        self.messages = []

    @cached_property
    def contents(self) -> dict:
        """Return defined API endpoint properties for the current type."""
        _properties = {}
        try:
            r = self.session.options(self.api_url, params={'show': 'readOnly'})
            # print(r.request.method, r.request.url, r.text)
            if r.ok:
                _properties = r.json()

        except (ConnectionError, ProxyError) as ex:
            Render.panel.failure(f'Failed getting types properties ({ex}).')

        return _properties

    @property
    def contents_updated(self) -> Generator:
        """Return the updated contents."""
        properties_comparison = []
        for field_name, field_data in self._contents_update(self.contents).items():
            field_data_ = field_data
            if isinstance(field_data_, dict):
                if 'data' in field_data_ and isinstance(field_data_['data'], list):
                    # some properties have a data key with an array of items.
                    field_data_ = field_data_['data'][0]

                    # if there is a data array, then the type should always be plural.
                    field_data_['type'] = Util.camel_string(field_data_['type']).plural()
            elif isinstance(field_data_, list):
                # in a few instance like attributeType the value of the properties key/value
                # pair is a list (currently the list only contains a single dict). to be safe
                # we loop over the list and update the type for each item.
                field_data_ = field_data_[0]
            else:
                Render.panel.failure(
                    'Invalid type properties data: field-name='
                    f'{field_name}, type={self.object_type}'
                )
            field_data_['name'] = field_name

            # capture input/output data for troubleshooting
            properties_comparison.append(
                {
                    'field_name': field_name,
                    'input_data': self.contents.get(field_name),
                    'output_data': field_data_,
                }
            )

            yield field_data_

        # write comparison data for troubleshooting
        self._write_comparison_data(properties_comparison)

    @property
    def content_models(self) -> Generator[PropertyModel, None, None]:
        """Return a list of PropertyModel objects."""
        for field_data in self.contents_updated:
            try:
                property_model = PropertyModel(**field_data)
            except ValidationError as ex:
                Render.panel.failure(
                    f'Failed generating property model: data={field_data} ({ex}).',
                )
            else:
                yield property_model

    @cached_property
    def properties_data_directory(self) -> Path:
        """Return the properties for the current type."""
        path = Path('tcex/api/tc/v3/_gen/properties_data')
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _contents_update(self, properties: dict) -> dict:
        """Update the properties contents, fixing issues in core data."""

        # add id field, if missing
        self._content_add_id(properties)

        # add security label field, if missing
        self._content_add_security_labels(properties)

        # add webLink field, if missing
        self._content_add_web_link(properties)

        # fix types
        self._content_fix_types(properties)

        # remove unused fields, if any
        self._content_remove_unused(properties)

        # update "bad" data
        self._content_update(properties)

        # maybe a temp issue?
        if self.object_type == 'indicators':
            properties['enrichment']['data'][0]['type'] = 'Enrichment'

        # critical fix for breaking API change
        if self.object_type in [
            'case_attributes',
            'group_attributes',
            'indicator_attributes',
            'victim_attributes',
        ]:
            self._content_update_fix_attribute_source_type(properties)

        if self.object_type == 'groups':
            self._content_update_fix_filename_max_length(properties)
            self._content_update_fix_down_vote_count(properties)
            self._content_update_fix_up_vote_count(properties)

        if self.object_type in [
            'cases',
            'tasks',
            'workflow_templates',
        ]:
            if properties.get('owner'):
                self._content_update_fix_read_only_owner(properties)
                self._content_update_fix_updatable_owner(properties)

            if properties.get('ownerId'):
                self._content_update_fix_read_only_owner_id(properties)
                self._content_update_fix_updatable_owner_id(properties)

        return properties

    def _content_update_fix_attribute_source_type(self, properties: dict):
        """."""
        title = 'Attribute Source Type'
        if properties['source']['type'] != 'String':
            properties['source']['type'] = 'String'
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
        else:
            self.messages.append(f'- [{self.object_type}] - {title} - fix NOT required.')

    def _content_update_fix_filename_max_length(self, properties: dict):
        """."""
        filename_max_length = 255
        title = 'Filename Max Length'
        if properties['fileName']['maxLength'] != filename_max_length:
            properties['fileName']['maxLength'] = filename_max_length
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
        else:
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix NOT required.')

    def _content_update_fix_down_vote_count(self, properties: dict):
        """."""
        title = 'Down Vote Count'
        if 'downVoteCount' in properties and properties['downVoteCount']['type'] == 'String':
            properties['downVoteCount']['type'] = 'Integer'
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
        else:
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix NOT required.')

    def _content_update_fix_up_vote_count(self, properties: dict):
        """."""
        title = 'Up Vote Count'
        if 'upVoteCount' in properties and properties['upVoteCount']['type'] == 'String':
            properties['upVoteCount']['type'] = 'Integer'
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
        else:
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix NOT required.')

    def _content_update_fix_read_only_owner(self, properties: dict):
        """."""
        title = 'Read Only Owner Fix'
        if properties['owner'].get('readOnly') is not True:
            properties['owner']['readOnly'] = True
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
        else:
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix NOT required.')

    def _content_update_fix_updatable_owner(self, properties: dict):
        """."""
        title = 'Updatable Owner Fix'
        if properties['owner'].get('updatable') is not False:
            properties['owner']['updatable'] = False
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
        else:
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix NOT required.')

    def _content_update_fix_read_only_owner_id(self, properties: dict):
        """."""
        title = 'Read Only Owner Id Fix'
        if properties['ownerId'].get('readOnly') is not True:
            properties['ownerId']['readOnly'] = True
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
        else:
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix NOT required.')

    def _content_update_fix_updatable_owner_id(self, properties: dict):
        """."""
        title = 'Updatable Owner Id Fix'
        if properties['ownerId'].get('updatable') is not False:
            properties['ownerId']['updatable'] = False
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix required.')
        else:
            self.messages.append(f'- [{self.object_type}] - ({title}) - fix NOT required.')

    def _content_add_id(self, properties: dict):
        """Add id field to properties.

        For some reason the core API does not return the id field for some types.
        """
        if 'id' not in properties:
            properties['id'] = {
                'required': False,
                'type': 'Integer',
                'description': 'The id of the **Object**',
                'readOnly': True,
                'updatable': False,
            }
            self.messages.append(f'- [{self.object_type}] - The id field is missing.')

    def _content_add_security_labels(self, properties: dict):
        """Add securityLabels field to properties."""
        if self.object_type in ['group_attributes', 'indicator_attributes']:
            if 'securityLabels' not in properties:
                properties['securityLabels'] = {
                    'data': [
                        {
                            'description': (
                                'A list of Security Labels corresponding to the Intel item '
                                '(NOTE: Setting this parameter will replace any existing '
                                'tag(s) with the one(s) specified)'
                            ),
                            'maxSize': 1000,
                            'required': False,
                            'type': 'SecurityLabel',
                        }
                    ]
                }
                self.messages.append(
                    f'- [{self.object_type}] - The securityLabels field is missing.'
                )
            else:
                self.messages.append(
                    f'- [{self.object_type}] - The securityLabels field is NOT missing.'
                )

    def _content_add_web_link(self, properties: dict):
        """Add webLink field to properties.

        For some reason the core API does not return the webLink field for some types.
        """
        if self.object_type in ['groups', 'indicators']:
            if 'webLink' not in properties:
                properties['webLink'] = {
                    'required': False,
                    'type': 'String',
                    'description': 'The object link.',
                    'readOnly': True,
                    'updatable': False,
                }
                self.messages.append(f'- [{self.object_type}] - The webLink field is missing.')
            else:
                self.messages.append(f'- [{self.object_type}] - The webLink field is NOT missing.')

    def _content_fix_types(self, _properties: dict):
        """Return modified type."""
        type_ = _properties.get('type')

        if type_ == 'attributes':
            if self.object_type == 'cases':
                _properties['type'] = 'caseAttributes'
            elif self.object_type == 'groups':
                _properties['type'] = 'groupAttributes'
            elif self.object_type == 'indicators':
                _properties['type'] = 'indicatorAttributes'
            elif self.object_type == 'victims':
                _properties['type'] = 'victimAttributes'

    def _content_remove_unused(self, properties: dict):
        """Remove unused fields from properties."""
        if self.object_type in ['attribute_types'] and 'owner' in properties:
            del properties['owner']

        if self.object_type in [
            'attribute_types',
            'case_attributes',
            'group_attributes',
            'indicator_attributes',
            'victim_attributes',
        ]:
            unused_fields = [
                'attributeTypeMappings',
                'ownerId',  # Core Issue: should not be listed for this type
                'ownerName',  # Core Issue: should not be listed for this type
                'settings',
                'tags',  # Core Issue: should not be listed for this type
                'webLink',  # Core Issue: should not be listed for this type
            ]
            for field in unused_fields:
                if field in properties:
                    del properties[field]

        if self.object_type in ['users']:
            unused_fields = [
                'customTqlTimeout',
                'disabled',
                'firstName',
                'jobFunction',
                'jobRole',
                'lastLogin',
                'lastName',
                'lastPasswordChange',
                'locked',
                'logoutIntervalMinutes',
                'owner',
                'ownerRoles',
                'password',
                'passwordResetRequired',
                'pseudonym',
                'systemRole',
                'termsAccepted',
                'termsAcceptedDate',
                'tqlTimeout',
                'twoFactorResetRequired',
                'uiTheme',
            ]
            for field in unused_fields:
                if field in properties:
                    del properties[field]

        if self.object_type in ['victims']:
            unused_fields = [
                'ownerId',  # Core Issue: should not be listed for this type
            ]
            for field in unused_fields:
                if field in properties:
                    del properties[field]

    def _content_update(self, properties: dict):
        """Update "bad" data in properties."""
        if self.object_type in ['groups']:
            # fixed fields that are missing readOnly property
            if 'downVoteCount' in properties:
                properties['downVoteCount']['readOnly'] = True
            if 'upVoteCount' in properties:
                properties['upVoteCount']['readOnly'] = True

        if self.object_type in ['victims']:
            # ownerName is readOnly, but readOnly is not defined in response from OPTIONS endpoint
            properties['ownerName']['readOnly'] = True

    def _write_comparison_data(self, properties_comparison: list):
        # write input/output data for troubleshooting
        output_file = self.properties_data_directory / f'{self.object_type}.json'
        with output_file.open(mode='w') as fh:
            json.dump(properties_comparison, fh, indent=2)



"""TcEx Framework Module"""

# standard library
import logging
import threading

_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class ExceptionThread(threading.Thread):
    """Thread that saves any uncaught exception into an instance variable for further inspection"""

    def __init__(self, *args, **kwargs):
        """Initialize thread"""
        super().__init__(*args, **kwargs)
        self.exception = None

    def run(self):
        """Run thread logic"""
        try:
            super().run()
        except Exception as ex:
            self.exception = ex
            _logger.exception(f'Unexpected exception occurred in thread with name: {self.name}')
            # let exception logic continue as normal
            raise



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class IntelRequirementModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='IntelRequirement Model',
    validate_assignment=True,
):
    """Intel_Requirement Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=False)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    associated_artifacts: 'ArtifactsModel' = Field(
        None,
        description='A list of Artifacts associated with this Group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedArtifacts',
    )
    associated_cases: 'CasesModel' = Field(
        None,
        description='A list of Cases associated with this Group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedCases',
    )
    associated_groups: 'GroupsModel' = Field(
        None,
        description='A list of groups associated with this group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedGroups',
    )
    associated_indicators: 'IndicatorsModel' = Field(
        None,
        description='A list of indicators associated with this group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedIndicators',
    )
    associated_victim_assets: 'VictimAssetsModel' = Field(
        None,
        description='A list of victim assets associated with this group.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedVictimAssets',
    )
    category: 'IntelReqTypeModel' = Field(
        None,
        description='The category of the intel requirement.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='category',
    )
    created_by: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The user who created the intel requirement.',
        read_only=True,
        title='createdBy',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the item was first created.',
        read_only=True,
        title='dateAdded',
    )
    description: str | None = Field(
        None,
        description='The description of the intel requirement.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='description',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    keyword_sections: list['KeywordSectionModel'] = Field(
        None,
        description='The section of the intel requirement that contains the keywords.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='keywordSections',
    )
    last_modified: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the Entity was last modified.',
        read_only=True,
        title='lastModified',
    )
    last_retrieved_date: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The last date the results were retrieved for the intel requirement.',
        read_only=True,
        title='lastRetrievedDate',
    )
    requirement_text: str | None = Field(
        None,
        description='The detailed text of the intel requirement.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='requirementText',
    )
    reset_results: bool = Field(
        None,
        description='Flag to reset results when updating keywords.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='resetResults',
    )
    results_link: str | None = Field(
        None,
        allow_mutation=False,
        description='A link to the results for the intel requirement.',
        read_only=True,
        title='resultsLink',
    )
    subtype: 'IntelReqTypeModel' = Field(
        None,
        description='The subtype of the intel requirement.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='subtype',
    )
    tags: 'TagsModel' = Field(
        None,
        description=(
            'A list of Tags corresponding to the item (NOTE: Setting this parameter will replace '
            'any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='tags',
    )
    unique_id: str | None = Field(
        None,
        description='The unique id of the intel requirement.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='uniqueId',
    )
    web_link: str | None = Field(
        None,
        allow_mutation=False,
        description='A link to the ThreatConnect details page for this entity.',
        read_only=True,
        title='webLink',
    )
    xid: str | None = Field(
        None,
        description='The xid of the item.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='xid',
    )

    @validator('associated_artifacts', always=True, pre=True)
    @classmethod
    def _validate_artifacts(cls, v):
        if not v:
            return ArtifactsModel()  # type: ignore
        return v

    @validator('associated_cases', always=True, pre=True)
    @classmethod
    def _validate_cases(cls, v):
        if not v:
            return CasesModel()  # type: ignore
        return v

    @validator('associated_groups', always=True, pre=True)
    @classmethod
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()  # type: ignore
        return v

    @validator('associated_indicators', always=True, pre=True)
    @classmethod
    def _validate_indicators(cls, v):
        if not v:
            return IndicatorsModel()  # type: ignore
        return v

    @validator('category', 'subtype', always=True, pre=True)
    @classmethod
    def _validate_intel_req_type(cls, v):
        if not v:
            return IntelReqTypeModel()  # type: ignore
        return v

    @validator('keyword_sections', always=True, pre=True)
    @classmethod
    def _validate_keyword_section(cls, v):
        if not v:
            return list['KeywordSectionModel']()  # type: ignore
        return v

    @validator('tags', always=True, pre=True)
    @classmethod
    def _validate_tags(cls, v):
        if not v:
            return TagsModel()  # type: ignore
        return v

    @validator('created_by', always=True, pre=True)
    @classmethod
    def _validate_user(cls, v):
        if not v:
            return UserModel()  # type: ignore
        return v

    @validator('associated_victim_assets', always=True, pre=True)
    @classmethod
    def _validate_victim_assets(cls, v):
        if not v:
            return VictimAssetsModel()  # type: ignore
        return v


class IntelRequirementDataModel(
    BaseModel,
    title='IntelRequirement Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Intel_Requirements Data Model"""

    data: list[IntelRequirementModel] | None = Field(
        [],
        description='The data for the IntelRequirements.',
        methods=['POST', 'PUT'],
        title='data',
    )


class IntelRequirementsModel(
    BaseModel,
    title='IntelRequirements Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Intel_Requirements Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[IntelRequirementModel] | None = Field(
        [],
        description='The data for the IntelRequirements.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactsModel
from tcex.api.tc.v3.cases.case_model import CasesModel
from tcex.api.tc.v3.groups.group_model import GroupsModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorsModel
from tcex.api.tc.v3.intel_requirements.intel_req_type_model import IntelReqTypeModel
from tcex.api.tc.v3.intel_requirements.keyword_sections.keyword_section_model import (
    KeywordSectionModel,
)
from tcex.api.tc.v3.security.users.user_model import UserModel
from tcex.api.tc.v3.tags.tag_model import TagsModel
from tcex.api.tc.v3.victim_assets.victim_asset_model import VictimAssetsModel

# add forward references
IntelRequirementDataModel.update_forward_refs()
IntelRequirementModel.update_forward_refs()
IntelRequirementsModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
import contextlib
import gzip
import hashlib
import json
import logging
import re
import shelve  # nosec
import sys
import time
import uuid
from collections import deque
from pathlib import Path
from typing import Any

# third-party
from requests import Session  # TYPE-CHECKING

# first-party
from tcex.api.tc.util.threat_intel_util import ThreatIntelUtil
from tcex.api.tc.v2.batch.association import Association
from tcex.api.tc.v2.batch.group import (
    Adversary,
    AttackPattern,
    Campaign,
    CourseOfAction,
    Document,
    Email,
    Event,
    Group,
    Incident,
    IntrusionSet,
    Malware,
    Report,
    Signature,
    Tactic,
    Threat,
    Tool,
    Vulnerability,
)
from tcex.api.tc.v2.batch.indicator import (
    ASN,
    CIDR,
    URL,
    Address,
    EmailAddress,
    File,
    Host,
    Indicator,
    Mutex,
    RegistryKey,
    UserAgent,
    custom_indicator_class_factory,
)
from tcex.input.input import Input  # TYPE-CHECKING
from tcex.logger.trace_logger import TraceLogger
from tcex.util import Util

# import local modules for dynamic reference
module = __import__(__name__)

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore

# define GroupType
GroupType = (
    Adversary
    | AttackPattern
    | Campaign
    | CourseOfAction
    | Document
    | Email
    | Event
    | Group
    | Incident
    | IntrusionSet
    | Malware
    | Report
    | Signature
    | Tactic
    | Threat
    | Tool
    | Vulnerability
)

# define IndicatorType
IndicatorType = (
    ASN
    | CIDR
    | URL
    | Address
    | EmailAddress
    | File
    | Host
    | Indicator
    | Mutex
    | RegistryKey
    | UserAgent
)


class BatchWriter:
    """ThreatConnect Batch Import Module

    Args:
        inputs: The App inputs.
        session_tc: The ThreatConnect API session.
        output_dir: The directory to write the batch JSON data.
    """

    def __init__(self, inputs: Input, session_tc: Session, output_dir: str, **kwargs):
        """Initialize instance properties."""
        self.inputs = inputs
        self.output_dir = output_dir
        self.output_extension = kwargs.get('output_extension')
        self.session_tc = session_tc
        self.write_callback = kwargs.get('write_callback')
        self.write_callback_kwargs = kwargs.get('write_callback_kwargs', {})

        # properties
        self._batch_files = []
        self._batch_max_chunk = 100_000
        self._batch_size = 0  # track current batch size
        self._batch_max_size = 75_000_000  # max size in bytes
        self.log = _logger
        self.tic = ThreatIntelUtil(self.session_tc)
        self.util = Util()

        # shelf settings
        self._group_shelf_fqfn = None
        self._indicator_shelf_fqfn = None

        # containers
        self._groups = None
        self._groups_shelf = None
        self._indicators = None
        self._indicators_shelf = None
        self._associations = None

        # build custom indicator classes
        self._gen_indicator_class()

    def _gen_indicator_class(self):  # pragma: no cover
        """Generate Custom Indicator Classes."""
        for entry in self.tic.indicator_types_data.values():
            name = entry['name']
            class_name = name.replace(' ', '')
            # temp fix for API issue where boolean are returned as strings
            entry['custom'] = self.util.to_bool(entry.get('custom') or False)

            if class_name in globals():
                # skip Indicator Type if a class already exists
                continue

            # Custom Indicator can have 3 values. Only add the value if it is set.
            value_fields = []
            if entry.get('value1Label'):
                value_fields.append(entry['value1Label'])
            if entry.get('value2Label'):
                value_fields.append(entry['value2Label'])
            if entry.get('value3Label'):
                value_fields.append(entry['value3Label'])
            value_count = len(value_fields)

            class_data = {}
            # Add Class for each Custom Indicator type to this module
            custom_class = custom_indicator_class_factory(name, Indicator, class_data, value_fields)
            setattr(module, class_name, custom_class)

            # Add Custom Indicator Method
            self._gen_indicator_method(name, custom_class, value_count)

    def _gen_indicator_method(
        self, name: str, custom_class: Any, value_count: int
    ):  # pragma: no cover
        """Dynamically generate custom Indicator methods.

        Args:
            name: The name of the method.
            custom_class: The class to add.
            value_count: The number of value parameters to support.
        """
        method_name = name.replace(' ', '_').lower()

        # Add Method for each Custom Indicator class
        def method_1(value1: str, xid, **kwargs):
            """Add Custom Indicator data to Batch"""
            indicator_obj = custom_class(value1, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        def method_2(value1: str, value2: str, xid, **kwargs):
            """Add Custom Indicator data to Batch"""
            indicator_obj = custom_class(value1, value2, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        def method_3(value1: str, value2: str, value3: str, xid, **kwargs):
            """Add Custom Indicator data to Batch"""
            indicator_obj = custom_class(value1, value2, value3, xid, **kwargs)
            return self._indicator(indicator_obj, kwargs.get('store', True))

        method = locals()[f'method_{value_count}']
        setattr(self, method_name, method)

    def _group(self, group_data: dict | GroupType, store: bool = True) -> dict | GroupType:
        """Return previously stored group or new group.

        Args:
            group_data: An Group dict or instance of a GroupType.
            store: If True the group data will be stored in instance list.
        """
        if store is False:
            return group_data

        # get xid from dict or from GroupType
        xid = group_data['xid'] if isinstance(group_data, dict) else group_data.xid

        if self.groups.get(xid) is not None:
            # return existing group from memory
            group_data = self.groups[xid]
        elif self.groups_shelf.get(xid) is not None:
            # return existing group from shelf
            group_data = self.groups_shelf[xid]
        else:
            # store new group
            self.groups[xid] = group_data

            # track total batch job data size as TI gets added
            if isinstance(group_data, dict):
                self._batch_size += sys.getsizeof(json.dumps(group_data))
            else:
                self._batch_size += sys.getsizeof(json.dumps(group_data.data))

            # max size hit, dump TI to disk
            if self._batch_size > self._batch_max_size:
                self.dump()
        return group_data

    def _indicator(
        self, indicator_data: dict | IndicatorType, store: bool = True
    ) -> dict | IndicatorType:
        """Return previously stored indicator or new indicator.

        Args:
            indicator_data: An Indicator dict or instance of IndicatorType.
            store: If True the indicator data will be stored in instance list.
        """
        if store is False:
            return indicator_data

        # get xid from dict or from IndicatorType
        xid = indicator_data['xid'] if isinstance(indicator_data, dict) else indicator_data.xid

        if self.indicators.get(xid) is not None:
            # return existing indicator from memory
            indicator_data = self.indicators[xid]
        elif self.indicators_shelf.get(xid) is not None:
            # return existing indicator from shelf
            indicator_data = self.indicators_shelf[xid]
        else:
            # store new indicators
            self.indicators[xid] = indicator_data

            # track total batch job data size as TI gets added
            if isinstance(indicator_data, dict):
                self._batch_size += sys.getsizeof(json.dumps(indicator_data))
            else:
                self._batch_size += sys.getsizeof(json.dumps(indicator_data.data))

            # max size hit, dump TI to disk
            if self._batch_size > self._batch_max_size:
                self.dump()
        return indicator_data

    @staticmethod
    def _indicator_values(indicator: str) -> list:
        """Process indicators expanding file hashes/custom indicators into multiple entries.

        Args:
            indicator: Indicator value represented as " : " delimited string.

        Returns:
            list: The list of indicators split on " : ".
        """
        indicator_list = [indicator]
        if indicator.count(' : ') > 0:
            # handle all multi-valued indicators types (file hashes and custom indicators)
            indicator_list = []

            # group 1 - lazy capture everything to first <space>:<space> or end of line
            iregx_pattern = r'^(.*?(?=\s\:\s|$))?'
            iregx_pattern += r'(?:\s\:\s)?'  # remove <space>:<space>
            # group 2 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead (optional <space>):<space> or end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=(?:\s)?\:\s|$))?'
            iregx_pattern += r'(?:(?:\s)?\:\s)?'  # remove (optional <space>):<space>
            # group 3 - look behind for <space>:<space>, lazy capture everything
            #           to look ahead end of line
            iregx_pattern += r'((?<=\s\:\s).*?(?=$))?$'
            iregx = re.compile(iregx_pattern)

            indicators = iregx.search(indicator)
            if indicators is not None:
                indicator_list = list(indicators.groups())

        return indicator_list

    def add_group(self, group_data: dict, **kwargs) -> dict | GroupType:
        """Add a group to Batch Job.

        .. code-block:: javascript

            {
                "name": "Example Incident",
                "type": "Incident",
                "attribute": [{
                    "type": "Description",
                    "displayed": false,
                    "value": "Example Description"
                }],
                "xid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904",
                "associatedGroupXid": [
                    "e336e2dd-5dfb-48cd-a33a-f8809e83e904:58",
                ],
                "tag": [{
                    "name": "China"
                }]
            }

        Args:
            group_data: The full Group data including attributes, labels, tags, and
                associations.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            (dict|GroupType): The new group dict/GroupType or the previously stored dict/GroupType.
        """
        return self._group(group_data, kwargs.get('store', True))

    def add_indicator(self, indicator_data: dict, **kwargs) -> dict | IndicatorType:
        """Add an indicator to Batch Job.

        .. code-block:: javascript

            {
                "type": "File",
                "rating": 5.00,
                "confidence": 50,
                "summary": "53c3609411c83f363e051d455ade78a7
                            : 57a49b478310e4313c54c0fee46e4d70a73dd580
                            : db31cb2a748b7e0046d8c97a32a7eb4efde32a0593e5dbd58e07a3b4ae6bf3d7",
                "associatedGroups": [
                    {
                        "groupXid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904"
                    }
                ],
                "attribute": [{
                    "type": "Source",
                    "displayed": true,
                    "value": "Malware Analysis provided by external AMA."
                }],
                "fileOccurrence": [{
                    "fileName": "drop1.exe",
                    "date": "2017-03-03T18:00:00-06:00"
                }],
                "tag": [{
                    "name": "China"
                }],
                "xid": "e336e2dd-5dfb-48cd-a33a-f8809e83e904:170139"
            }

        Args:
            indicator_data: The Full Indicator data including attributes, labels, tags,
                and associations.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            (dict|IndicatorType): The new group dict/IndicatorType
            or the previously stored dict/IndicatorType.
        """
        if indicator_data.get('type') not in ['Address', 'EmailAddress', 'File', 'Host', 'URL']:
            # for custom indicator types the valueX fields are required.
            # using the summary we can build the values
            index = 1
            for value in self._indicator_values(indicator_data['summary']):
                indicator_data[f'value{index}'] = value
                index += 1

        if indicator_data.get('type') == 'File':
            # convert custom field name to the appropriate value for batch v2
            size = indicator_data.pop('size', None)
            if size is not None:
                indicator_data['intValue1'] = size

        if indicator_data.get('type') == 'Host':
            # convert custom field name to the appropriate value for batch v2
            dns_active = indicator_data.pop('dnsActive', None)

            if dns_active is not None:
                indicator_data['flag1'] = dns_active
            whois_active = indicator_data.pop('whoisActive', None)

            if whois_active is not None:
                indicator_data['flag2'] = whois_active

        return self._indicator(indicator_data, kwargs.get('store', True))

    def add_association(self, association: Association | dict) -> Association:
        """Add an association to Batch Job."""
        if isinstance(association, dict):
            association = Association(**association)
        self.associations.add(association)
        return association

    def address(self, ip: str, **kwargs) -> Address | dict:
        """Add Address data to Batch.

        Args:
            ip: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Address: An instance of the Address class.
        """
        indicator_obj = Address(ip, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def adversary(self, name: str, **kwargs) -> Adversary:
        """Add Adversary data to Batch.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Adversary: An instance of the Adversary class.
        """
        group_obj = Adversary(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def attack_pattern(self, name: str, **kwargs) -> AttackPattern:
        """Add Attack Pattern data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            AttackPattern: An instance of the AttackPattern class.
        """
        group_obj = AttackPattern(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def asn(self, as_number: str, **kwargs) -> ASN:
        """Add ASN data to Batch.

        Args:
            as_number: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            ASN: An instance of the ASN class.
        """
        indicator_obj = ASN(as_number, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def campaign(self, name: str, **kwargs) -> Campaign:
        """Add Campaign data to Batch.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            first_seen (str, kwargs): The first seen datetime expression for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Campaign: An instance of the Campaign class.
        """
        group_obj = Campaign(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def cidr(self, block: str, **kwargs) -> CIDR:
        """Add CIDR data to Batch.

        Args:
            block: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            CIDR: An instance of the CIDR class.
        """
        indicator_obj = CIDR(block, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def close(self):
        """Cleanup batch job."""
        self.dump()

        # cleanup shelf files
        try:
            self.groups_shelf.close()
            self.group_shelf_fqfn.unlink()
        except Exception as ex:
            self.log.warning(f'action=batch-close, filename={self.group_shelf_fqfn} exception={ex}')

        # cleanup shelf files
        try:
            self.indicators_shelf.close()
            self.indicator_shelf_fqfn.unlink()
        except Exception as ex:
            self.log.warning(
                f'action=batch-close, filename={self.indicator_shelf_fqfn} exception={ex}'
            )

    def course_of_action(self, name: str, **kwargs) -> CourseOfAction:
        """Add Course Of Action Pattern data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            CourseOfAction: An instance of the CourseOfAction class.
        """
        group_obj = CourseOfAction(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    @property
    def data(self) -> dict:
        """Return the batch indicator/group and file data to be sent to the ThreatConnect API.

        **Processing Order:**
        * Process groups in memory up to max batch size.
        * Process groups in shelf to max batch size.
        * Process indicators in memory up to max batch size.
        * Process indicators in shelf up to max batch size.

        This method will remove the group/indicator from memory and/or shelf.

        Returns:
            dict: A dictionary of group, indicators, and/or file data.
        """
        data = {'file': {}, 'group': [], 'indicator': []}
        tracker = {'count': 0}

        # process group from memory, returning if max values have been reached
        if self.data_groups(data, self.groups, tracker) is True:
            return data

        # process group from shelf file, returning if max values have been reached
        if self.data_groups(data, self.groups_shelf, tracker) is True:
            return data

        # process indicator from memory, returning if max values have been reached
        if self.data_indicators(data, self.indicators, tracker) is True:
            return data

        # process indicator from shelf file, returning if max values have been reached
        if self.data_indicators(data, self.indicators_shelf, tracker) is True:
            return data

        return data

    def data_group_association(self, data: dict, tracker: dict, xid: str):
        """Return group dict array following all associations.

        The *data* dict is passed by reference to make it easier to update both the group data
        and file data inline versus passing the data all the way back up to the calling methods.

        Args:
            data: The data dict to update with group and file data.
            tracker: A dict containing total count of all entities collected.
            xid: The xid of the group to retrieve associations.
        """
        xids = deque()
        xids.append(xid)

        while xids:
            xid = xids.popleft()  # remove current xid
            group_data = None

            if xid in self.groups:
                group_data = self.groups.get(xid)
                del self.groups[xid]
            elif xid in self.groups_shelf:
                group_data = self.groups_shelf.get(xid)
                del self.groups_shelf[xid]

            if group_data:
                file_data, group_data = self.data_group_type(group_data)
                data['group'].append(group_data)
                if file_data:
                    data['file'][xid] = file_data

                # update entity trackers
                tracker['count'] += 1

                # extend xids with any groups associated with the same GroupType
                xids.extend(group_data.get('associatedGroupXid', []))

    @staticmethod
    def data_group_type(group_data: dict | GroupType) -> tuple[dict, dict]:
        """Return dict representation of group data and file data.

        Args:
            group_data: The group data.

        Returns:
            Tuple[dict, dict]: A tuple containing file_data and group_data.
        """
        file_data = {}
        if isinstance(group_data, dict):
            # process file content
            file_content = group_data.pop('fileContent', None)
            if file_content is not None:
                file_data = {
                    'fileContent': file_content,
                    'fileName': group_data.get('fileName'),
                    'type': group_data.get('type'),
                }
        else:
            # get the file data from the GroupType and return dict format of GroupType
            if group_data.data.get('type') in ['Document', 'Report']:
                file_data = group_data.file_data
            group_data = group_data.data

        return file_data, group_data

    def data_groups(self, data: dict, groups: dict | shelve.Shelf[Any], tracker: dict) -> bool:
        """Process Group data.

        Args:
            data: The data dict to update with group and file data.
            groups: The list of groups to process.
            tracker: A dict containing total count of all entities collected.

        Returns:
            bool: True if max values have been hit, else False.
        """
        # convert groups.keys() to a list to prevent dictionary change error caused by
        # the data_group_association function deleting items from the GroupType.

        # process the group
        for xid in list(groups.keys()):
            # get association from group data
            self.data_group_association(data, tracker, xid)

            if tracker['count'] % 10_000 == 0:
                # log count/size at a sane level
                self.log.debug(f'feature=batch, action=data-groups, count={tracker.get("count"):,}')

        return False

    def data_indicators(
        self, data: dict, indicators: dict | shelve.Shelf[Any], tracker: dict
    ) -> bool:
        """Process Indicator data.

        Args:
            data: The data dict to update with group and file data.
            indicators: The list of indicators to process.
            tracker: A dict containing total count of all entities collected.

        Returns:
            bool: True if max values have been hit, else False.
        """
        # process the indicator
        for xid, indicator_data in list(indicators.items()):
            if not isinstance(indicator_data, dict):
                data['indicator'].append(indicator_data.data)
            else:
                data['indicator'].append(indicator_data)
            del indicators[xid]

            # update entity trackers
            tracker['count'] += 1

            if tracker['count'] % 10_000 == 0:
                # log count/size at a sane level
                self.log.debug(
                    """feature=batch, action=data-indicators, """
                    f"""count={tracker.get('count'):,}"""
                )
        return False

    def document(self, name: str, file_name: str, **kwargs) -> Document:
        """Add Document data to Batch.

        Args:
            name: The name for this Group.
            file_name: The name for the attached file for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            file_content (str;method, kwargs): The file contents or
                callback method to retrieve file content.
            malware (bool, kwargs): If true the file is considered malware.
            password (bool, kwargs): If malware is true a password for the zip archive is
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Document: An instance of the Document class.
        """
        group_obj = Document(name, file_name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def dump(self):
        """Process Batch request to ThreatConnect API."""
        content = self.data
        content.pop('file', {})
        if not content.get('group') and not content.get('indicator'):
            return

        # special code for debugging App using batchV2.
        self.write_batch_json(content)

        # store the length of the batch data to use for poll interval calculations
        self.log.info(f'feature=batch, event=dump, type=group, count={len(content["group"]):,}')
        self.log.info(
            f'feature=batch, event=dump, type=indicator, count={len(content["indicator"]):,}'
        )
        self.log.info(f'feature=batch, event=dump, type=batch, size={self._batch_size:,}')

        # reset batch size after dump
        self._batch_size = 0

    def email(self, name: str, subject: str, header: str, body: str, **kwargs) -> Email:
        """Add Email data to Batch.

        Args:
            name: The name for this Group.
            subject: The subject for this Email.
            header: The header for this Email.
            body: The body for this Email.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            from_addr (str, kwargs): The **from** address for this Email.
            to_addr (str, kwargs): The **to** address for this Email.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Email: An instance of the Email class.
        """
        group_obj = Email(name, subject, header, body, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def email_address(self, address: str, **kwargs) -> EmailAddress:
        """Add Email Address data to Batch.

        Args:
            address: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            EmailAddress: An instance of the EmailAddress class.
        """
        indicator_obj = EmailAddress(address, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def event(self, name: str, **kwargs) -> Event:
        """Add Event data to Batch.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            event_date (str, kwargs): The event datetime expression for this Group.
            status (str, kwargs): The status for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Event: An instance of the Event class.
        """
        group_obj = Event(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def file(
        self,
        md5: str | None = None,
        sha1: str | None = None,
        sha256: str | None = None,
        **kwargs,
    ) -> File:
        """Add File data to Batch.

        .. note:: A least one file hash value must be specified.

        Args:
            md5: The md5 value for this Indicator.
            sha1: The sha1 value for this Indicator.
            sha256: The sha256 value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            size (str, kwargs): The file size for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            File: An instance of the File class.

        """
        indicator_obj = File(md5, sha1, sha256, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    @staticmethod
    def generate_xid(identifier: list | str | None = None) -> str:
        """Generate xid from provided identifiers.

        .. Important::  If no identifier is provided a unique xid will be returned, but it will
                        not be reproducible. If a list of identifiers are provided they must be
                        in the same order to generate a reproducible xid.

        Args:
            identifier:  Value(s) to be used to make a unique and reproducible xid.

        """
        if identifier is None:
            identifier = str(uuid.uuid4())
        elif isinstance(identifier, list):
            identifier = '-'.join([str(i) for i in identifier])
            # IMPORTANT: Do not remove this duplicate line or xids will change.
            identifier = hashlib.sha256(identifier.encode('utf-8')).hexdigest()
        return hashlib.sha256(identifier.encode('utf-8')).hexdigest()

    def group(self, group_type: str, name: str, **kwargs) -> GroupType | dict:
        """Add Group data to Batch.

        Args:
            group_type: The ThreatConnect define Group type.
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            GroupType: An instance of one of the Group classes.
        """
        group_obj = Group(group_type, name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))

    @property
    def group_shelf_fqfn(self) -> Path:
        """Return groups shelf fully qualified filename.

        For testing/debugging a previous shelf file can be copied into the tc_temp_path directory
        instead of creating a new shelf file.
        """
        if self._group_shelf_fqfn is None:
            # new shelf file
            self._group_shelf_fqfn = self.inputs.model.tc_temp_path / f'groups-{uuid.uuid4()!s}'
        return self._group_shelf_fqfn

    @property
    def associations(self) -> dict:
        """Return dictionary of all Associations data."""
        if self._associations is None:
            self._associations = set()
        return self._associations

    @property
    def groups(self) -> dict:
        """Return dictionary of all Groups data."""
        if self._groups is None:
            self._groups = {}
        return self._groups

    @property
    def groups_shelf(self) -> shelve.Shelf[Any]:
        """Return dictionary of all Groups data."""
        if self._groups_shelf is None:
            self._groups_shelf = (
                shelve.open(self.group_shelf_fqfn, writeback=False)  # nosec  # noqa: SIM115
            )
        return self._groups_shelf

    def host(self, hostname: str, **kwargs) -> Host:
        """Add Host data to Batch.

        Args:
            hostname: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            dns_active (bool, kwargs): If True DNS active is enabled for this indicator.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            whois_active (bool, kwargs): If True WhoIs active is enabled for this indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Host: An instance of the Host class.
        """
        indicator_obj = Host(hostname, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def incident(self, name: str, **kwargs) -> Incident:
        """Add Incident data to Batch.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            event_date (str, kwargs): The event datetime expression for this Group.
            status (str, kwargs): The status for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Incident: An instance of the Incident class.
        """
        group_obj = Incident(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def indicator(self, indicator_type: str, summary: str, **kwargs) -> IndicatorType:
        """Add Indicator data to Batch.

        Args:
            indicator_type: The ThreatConnect define Indicator type.
            summary: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            IndicatorType: An instance of one of the Indicator classes.
        """
        indicator_obj = Indicator(indicator_type, summary, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    @property
    def indicator_shelf_fqfn(self) -> Path:
        """Return indicator shelf fully qualified filename.

        For testing/debugging a previous shelf file can be copied into the tc_temp_path directory
        instead of creating a new shelf file.
        """
        if self._indicator_shelf_fqfn is None:
            # new shelf file
            self._indicator_shelf_fqfn = (
                self.inputs.model.tc_temp_path / f'indicators-{uuid.uuid4()!s}'
            )
        return self._indicator_shelf_fqfn

    @property
    def indicators(self) -> dict:
        """Return dictionary of all Indicator data."""
        if self._indicators is None:
            # plain dict, but could be something else in future
            self._indicators = {}
        return self._indicators

    @property
    def indicators_shelf(self) -> shelve.Shelf[Any]:
        """Return dictionary of all Indicator data."""
        if self._indicators_shelf is None:
            self._indicators_shelf = shelve.open(  # nosec  # noqa: SIM115
                self.indicator_shelf_fqfn, writeback=False
            )
        return self._indicators_shelf

    def intrusion_set(self, name: str, **kwargs) -> IntrusionSet:
        """Add Intrusion Set data to Batch.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            IntrusionSet: An instance of the IntrusionSet class.
        """
        group_obj = IntrusionSet(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def malware(self, name: str, **kwargs) -> Malware:
        """Add Malware data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Malware: An instance of the Malware class.
        """
        group_obj = Malware(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def mutex(self, mutex: str, **kwargs) -> Mutex:
        """Add Mutex data to Batch.

        Args:
            mutex: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Mutex: An instance of the Mutex class.
        """
        indicator_obj = Mutex(mutex, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def registry_key(
        self, key_name: str, value_name: str, value_type: str, **kwargs
    ) -> RegistryKey:
        """Add Registry Key data to Batch.

        Args:
            key_name: The key_name value for this Indicator.
            value_name: The value_name value for this Indicator.
            value_type: The value_type value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            RegistryKey: An instance of the Registry Key class.
        """
        indicator_obj = RegistryKey(key_name, value_name, value_type, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def report(self, name: str, **kwargs) -> Report:
        """Add Report data to Batch.

        Args:
            name: The name for this Group.
            file_name (str): The name for the attached file for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            file_content (str;method, kwargs): The file contents or callback method to retrieve
                file content.
            publish_date (str, kwargs): The publish datetime expression for this Group.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Report: An instance of the Report class.
        """
        group_obj = Report(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def save(self, resource: dict | GroupType | IndicatorType):
        """Save group|indicator dict, GroupType, or IndicatorTypes to shelve.

        Best effort to save group/indicator data to disk.  If for any reason the save fails
        the data will still be accessible from list in memory.

        Args:
            resource: The Group or Indicator data.
        """
        resource_type = None
        xid = None
        if isinstance(resource, dict):
            resource_type = resource.get('type')
            xid = resource.get('xid')
        else:
            resource_type = resource.type
            xid = resource.xid

        if resource_type is not None and xid is not None:
            saved = True

            if resource_type in self.tic.group_types:
                try:
                    # groups
                    self.groups_shelf[xid] = resource
                except Exception:
                    saved = False

                if saved:
                    with contextlib.suppress(KeyError):
                        # if group was saved twice it would already be delete
                        del self.groups[xid]
            elif resource_type in self.tic.indicator_types_data:
                try:
                    # indicators
                    self.indicators_shelf[xid] = resource
                except Exception:
                    saved = False

                if saved:
                    with contextlib.suppress(KeyError):
                        # if indicator was saved twice it would already be delete
                        del self.indicators[xid]

    def signature(
        self, name: str, file_name: str, file_type: str, file_text: str, **kwargs
    ) -> Signature:
        """Add Signature data to Batch.

        Valid file_types:
        + Snort ®
        + Suricata
        + YARA
        + ClamAV ®
        + OpenIOC
        + CybOX ™
        + Bro
        + Regex
        + SPL - Splunk ® Search Processing Language

        Args:
            name: The name for this Group.
            file_name: The name for the attached signature for this Group.
            file_type: The signature type for this Group.
            file_text: The signature content for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Signature: An instance of the Signature class.
        """
        group_obj = Signature(name, file_name, file_type, file_text, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def tactic(self, name: str, **kwargs) -> Tactic:
        """Add Tactic data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Tactic: An instance of the Tactic class.
        """
        group_obj = Tactic(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def threat(self, name: str, **kwargs) -> Threat:
        """Add Threat data to Batch.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Threat: An instance of the Threat class.
        """
        group_obj = Threat(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def tool(self, name: str, **kwargs) -> Tool:
        """Add Tool data to Batch object.

        Args:
            name: The name for this Group.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            Tool: An instance of the Tool class.
        """
        group_obj = Tool(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def user_agent(self, text: str, **kwargs) -> UserAgent:
        """Add User Agent data to Batch.

        Args:
            text: The value for this Indicator.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.
            **kwargs: Additional keyword arguments.

        Returns:
            UserAgent: An instance of the UserAgent class.
        """
        indicator_obj = UserAgent(text, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def url(self, text: str, **kwargs) -> URL:
        """Add URL Address data to Batch.

        Args:
            text: The value for this Indicator.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): The date timestamp the Indicator was created.
            last_modified (str, kwargs): The date timestamp the Indicator was last modified.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            URL: An instance of the URL class.
        """
        indicator_obj = URL(text, **kwargs)
        return self._indicator(indicator_obj, kwargs.get('store', True))  # type: ignore

    def vulnerability(self, name: str, **kwargs) -> Vulnerability:
        """Add Vulnerability data to Batch.

        Args:
            name: The name for this Group.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            date_added (str, kwargs): The date timestamp the Indicator was created.
            xid (str, kwargs): The external id for this Group.
            store: (bool, kwargs): Advanced - Defaults to True. If True
                the indicator data will be stored in instance list.

        Returns:
            Vulnerability: An instance of the Vulnerability class.
        """
        group_obj = Vulnerability(name, **kwargs)
        return self._group(group_obj, kwargs.get('store', True))  # type: ignore

    def write_batch_json(self, content: dict):
        """Write batch json data to a file."""
        if content:
            # get timestamp as a string without decimal place and consistent length
            filename = f'{round(time.time() * 10000000)!s}.json.gz'
            if self.output_extension is not None:
                # add any additional extension provided
                filename += self.output_extension
            # TODO: is this needed
            self._batch_files.append(filename)
            fqfn = Path(self.output_dir) / filename
            with gzip.open(fqfn, mode='wt', encoding='utf-8') as fh:
                json.dump(content, fh)

            # send callback the filename
            if callable(self.write_callback):
                self.write_callback(fqfn, **self.write_callback_kwargs)



"""TcEx Framework Module"""

# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v2.threat_intelligence.mapping.group.group import Group

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import ThreatIntelligence


class Tactic(Group):
    """Unique API calls for Tactic API Endpoints

    Args:
        ti (ThreatIntelligence): An instance of the ThreatIntelligence Class.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
    """

    def __init__(self, ti: 'ThreatIntelligence', **kwargs):
        """Initialize instance properties."""
        super().__init__(ti, sub_type='Tactic', api_entity='tactic', api_branch='tactics', **kwargs)



"""TcEx Framework Module"""

# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class UserFilter(FilterABC):
    """Filter Object for Users"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.USERS.value

    def disabled(self, operator: Enum, disabled: bool):
        """Filter Disabled based on **disabled** keyword.

        Args:
            operator: The operator enum for the filter.
            disabled: A flag indicating whether or not the user's account has been disabled.
        """
        self._tql.add_filter('disabled', operator, disabled, TqlType.BOOLEAN)

    def first_name(self, operator: Enum, first_name: list | str):
        """Filter First Name based on **firstName** keyword.

        Args:
            operator: The operator enum for the filter.
            first_name: The first name of the user.
        """
        if isinstance(first_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('firstName', operator, first_name, TqlType.STRING)

    def group_id(self, operator: Enum, group_id: int | list):
        """Filter groupID based on **groupId** keyword.

        Args:
            operator: The operator enum for the filter.
            group_id: The ID of the group the user belongs to.
        """
        if isinstance(group_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('groupId', operator, group_id, TqlType.INTEGER)

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the user.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def job_function(self, operator: Enum, job_function: list | str):
        """Filter Job Function based on **jobFunction** keyword.

        Args:
            operator: The operator enum for the filter.
            job_function: The user's job function.
        """
        if isinstance(job_function, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('jobFunction', operator, job_function, TqlType.STRING)

    def job_role(self, operator: Enum, job_role: list | str):
        """Filter Job Role based on **jobRole** keyword.

        Args:
            operator: The operator enum for the filter.
            job_role: The user's job role.
        """
        if isinstance(job_role, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('jobRole', operator, job_role, TqlType.STRING)

    def last_login(self, operator: Enum, last_login: Arrow | datetime | int | str):
        """Filter Last Login based on **lastLogin** keyword.

        Args:
            operator: The operator enum for the filter.
            last_login: The last time the user logged in.
        """
        last_login = self.util.any_to_datetime(last_login).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastLogin', operator, last_login, TqlType.STRING)

    def last_name(self, operator: Enum, last_name: list | str):
        """Filter Last Name based on **lastName** keyword.

        Args:
            operator: The operator enum for the filter.
            last_name: The last name of the user.
        """
        if isinstance(last_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('lastName', operator, last_name, TqlType.STRING)

    def last_password_change(
        self, operator: Enum, last_password_change: Arrow | datetime | int | str
    ):
        """Filter Last Password Change based on **lastPasswordChange** keyword.

        Args:
            operator: The operator enum for the filter.
            last_password_change: The last time the user changed their password.
        """
        last_password_change = self.util.any_to_datetime(last_password_change).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('lastPasswordChange', operator, last_password_change, TqlType.STRING)

    def locked(self, operator: Enum, locked: bool):
        """Filter Locked based on **locked** keyword.

        Args:
            operator: The operator enum for the filter.
            locked: A flag indicating whether or not the user's account has been locked.
        """
        self._tql.add_filter('locked', operator, locked, TqlType.BOOLEAN)

    def password_reset_required(self, operator: Enum, password_reset_required: bool):
        """Filter Password Reset Required based on **passwordResetRequired** keyword.

        Args:
            operator: The operator enum for the filter.
            password_reset_required: A flag indicating whether or not the user's password needs to
                be reset upoin next login.
        """
        self._tql.add_filter(
            'passwordResetRequired', operator, password_reset_required, TqlType.BOOLEAN
        )

    def pseudonym(self, operator: Enum, pseudonym: list | str):
        """Filter Pseudonym based on **pseudonym** keyword.

        Args:
            operator: The operator enum for the filter.
            pseudonym: The user's pseudonym.
        """
        if isinstance(pseudonym, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('pseudonym', operator, pseudonym, TqlType.STRING)

    def system_role(self, operator: Enum, system_role: list | str):
        """Filter Role Name based on **systemRole** keyword.

        Args:
            operator: The operator enum for the filter.
            system_role: The system role name defined for the user.
        """
        if isinstance(system_role, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('systemRole', operator, system_role, TqlType.STRING)

    def terms_accepted(self, operator: Enum, terms_accepted: bool):
        """Filter Terms Accepted based on **termsAccepted** keyword.

        Args:
            operator: The operator enum for the filter.
            terms_accepted: Flag indicating whether user has accepted the current Terms of Service.
        """
        self._tql.add_filter('termsAccepted', operator, terms_accepted, TqlType.BOOLEAN)

    def terms_accepted_date(
        self, operator: Enum, terms_accepted_date: Arrow | datetime | int | str
    ):
        """Filter Terms Accepted Date based on **termsAcceptedDate** keyword.

        Args:
            operator: The operator enum for the filter.
            terms_accepted_date: Date and time the user accepted the current Terms of Service.
        """
        terms_accepted_date = self.util.any_to_datetime(terms_accepted_date).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('termsAcceptedDate', operator, terms_accepted_date, TqlType.STRING)

    def tql_timeout(self, operator: Enum, tql_timeout: int | list):
        """Filter TQL Timeout based on **tqlTimeout** keyword.

        Args:
            operator: The operator enum for the filter.
            tql_timeout: The custom TQL timeout value (if defined).
        """
        if isinstance(tql_timeout, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('tqlTimeout', operator, tql_timeout, TqlType.INTEGER)

    def user_name(self, operator: Enum, user_name: list | str):
        """Filter User Name based on **userName** keyword.

        Args:
            operator: The operator enum for the filter.
            user_name: The user name of the user.
        """
        if isinstance(user_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('userName', operator, user_name, TqlType.STRING)



"""TcEx Framework Module"""

# third-party
from pydantic import BaseModel, Field, validator

# first-party
from tcex.app.config.install_json import InstallJson
from tcex.input.field_type.sensitive import Sensitive


class ApiModel(BaseModel):
    """API Model

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * Organization
    * TriggerService
    * WebhookTriggerService
    """

    api_default_org: str | None = Field(
        None,
        description='The default ThreatConnect Org for the current API user.',
        inclusion_reason='runtimeLevel',
    )
    # alternate authentication credential when tc_token is not passed
    tc_api_access_id: str | None = Field(
        None,
        description='A ThreatConnect API Access Id.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_api_path: str = Field(
        'https://api.threatconnect.com',
        description='The URL for the ThreatConnect API.',
        inclusion_reason='runtimeLevel',
    )
    # alternate authentication credential when tc_token is not passed
    tc_api_secret_key: Sensitive | None = Field(
        None,
        description='A ThreatConnect API Secret Key.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_log_curl: bool = Field(
        default=False,
        description='Flag to enable logging curl commands.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_token: Sensitive | None = Field(
        None,
        description='A ThreatConnect API token.',
        inclusion_reason='runtimeLevel',
    )
    tc_token_expires: int | None = Field(
        None,
        description='The expiration timestamp in epoch for tc_token.',
        inclusion_reason='runtimeLevel',
    )
    tc_verify: bool = Field(
        default=True,
        description='Flag to enable SSL validation for API requests.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )

    @validator('tc_token', always=True, pre=True)
    @classmethod
    def one_set_of_credentials(cls, v, values):
        """Validate that one set of credentials is provided for the TC API."""
        _ij = InstallJson()

        # external Apps: require credentials and would not have an install.json file
        # organization (job) Apps: require credentials
        # playbook Apps: require credentials
        # service Apps: get token on createConfig message or during request
        if (
            (
                _ij.fqfn.is_file() is False
                or (_ij.model.is_playbook_app or _ij.model.is_organization_app)
            )
            and v is None
            and not all([values.get('tc_api_access_id'), values.get('tc_api_secret_key')])
        ):
            ex_msg = (
                'At least one set of ThreatConnect credentials must be provided '
                '(tc_api_access_id/tc_api_secret key OR tc_token/tc_token_expires).'
            )
            raise ValueError(ex_msg)
        return v



"""TcEx Framework Module"""

# flake8:noqa
from .key_value_api import KeyValueApi
from .key_value_mock import KeyValueMock
from .key_value_redis import KeyValueRedis
from .redis_client import RedisClient

__all__ = ['KeyValueApi', 'KeyValueMock', 'KeyValueRedis', 'RedisClient']



"""TcEx Framework Module"""

# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class TaskModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Task Model',
    validate_assignment=True,
):
    """Task Model"""

    _associated_type = PrivateAttr(default=False)
    _cm_type = PrivateAttr(default=True)
    _shared_type = PrivateAttr(default=False)
    _staged = PrivateAttr(default=False)

    artifacts: 'ArtifactsModel' = Field(
        None,
        description='A list of Artifacts corresponding to the Task.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='artifacts',
    )
    assignee: 'AssigneeModel' = Field(
        None,
        description='The user or group Assignee object for the Task.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='assignee',
    )
    case_id: int | None = Field(
        None,
        description='The **case id** for the Task.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseXid',
        title='caseId',
    )
    case_xid: str | None = Field(
        None,
        description='The **case xid** for the Task.',
        methods=['POST'],
        read_only=False,
        required_alt_field='caseId',
        title='caseXid',
    )
    completed_by: str | None = Field(
        None,
        allow_mutation=False,
        description='The **completed by** for the Task.',
        read_only=True,
        title='completedBy',
    )
    completed_date: datetime | None = Field(
        None,
        description='The completion date of the Task.',
        methods=['POST'],
        read_only=False,
        title='completedDate',
    )
    config_playbook: str | None = Field(
        None,
        allow_mutation=False,
        description='The **config playbook** for the Task.',
        read_only=True,
        title='configPlaybook',
    )
    config_task: dict | list[dict] | None = Field(
        None,
        allow_mutation=False,
        description='The **config task** for the Task.',
        read_only=True,
        title='configTask',
    )
    dependent_on_id: int | None = Field(
        None,
        description='The ID of another Task that this Task is dependent upon.',
        methods=['POST'],
        read_only=False,
        title='dependentOnId',
    )
    description: str | None = Field(
        None,
        description='The **description** for the Task.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='description',
    )
    due_date: datetime | None = Field(
        None,
        description='The due date of the Task.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='dueDate',
    )
    duration: int | None = Field(
        None,
        allow_mutation=False,
        description='The **duration** for the Task.',
        read_only=True,
        title='duration',
    )
    duration_type: str | None = Field(
        None,
        description='The **duration type** for the Task.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='durationType',
    )
    id: int | None = Field(  # type: ignore
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        description='The **name** for the Task.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='name',
    )
    notes: 'NotesModel' = Field(
        None,
        description='A list of Notes corresponding to the Task.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='notes',
    )
    owner: str | None = Field(
        None,
        allow_mutation=False,
        description='The name of the Owner of the Case.',
        read_only=True,
        title='owner',
    )
    parent_case: 'CaseModel' = Field(
        None,
        allow_mutation=False,
        description='The **parent case** for the Task.',
        read_only=True,
        title='parentCase',
    )
    required: bool = Field(
        None,
        description='Flag indicating whether or not the task is required.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='required',
    )
    status: str | None = Field(
        None,
        description='The **status** for the Task.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='status',
    )
    workflow_phase: int | None = Field(
        None,
        description='The phase of the workflow.',
        methods=['POST'],
        read_only=False,
        title='workflowPhase',
    )
    workflow_step: int | None = Field(
        None,
        description='The step of the workflow.',
        methods=['POST'],
        read_only=False,
        title='workflowStep',
    )
    xid: str | None = Field(
        None,
        description='The **xid** for the Task.',
        methods=['POST'],
        read_only=False,
        title='xid',
    )

    @validator('artifacts', always=True, pre=True)
    @classmethod
    def _validate_artifacts(cls, v):
        if not v:
            return ArtifactsModel()  # type: ignore
        return v

    @validator('assignee', always=True, pre=True)
    @classmethod
    def _validate_assignee(cls, v):
        if not v:
            return AssigneeModel()  # type: ignore
        return v

    @validator('parent_case', always=True, pre=True)
    @classmethod
    def _validate_case(cls, v):
        if not v:
            return CaseModel()  # type: ignore
        return v

    @validator('notes', always=True, pre=True)
    @classmethod
    def _validate_notes(cls, v):
        if not v:
            return NotesModel()  # type: ignore
        return v


class TaskDataModel(
    BaseModel,
    title='Task Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Tasks Data Model"""

    data: list[TaskModel] | None = Field(
        [],
        description='The data for the Tasks.',
        methods=['POST', 'PUT'],
        title='data',
    )


class TasksModel(
    BaseModel,
    title='Tasks Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Tasks Model"""

    _mode_support = PrivateAttr(default=False)

    data: list[TaskModel] | None = Field(
        [],
        description='The data for the Tasks.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactsModel
from tcex.api.tc.v3.cases.case_model import CaseModel
from tcex.api.tc.v3.notes.note_model import NotesModel
from tcex.api.tc.v3.security.assignee_model import AssigneeModel

# add forward references
TaskDataModel.update_forward_refs()
TaskModel.update_forward_refs()
TasksModel.update_forward_refs()



"""TcEx Framework Module"""

# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql import Tql
from tcex.api.tc.v3.tql.tql_operator import TqlOperator
from tcex.api.tc.v3.tql.tql_type import TqlType


class ResultFilter(FilterABC):
    """Filter Object for Results"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.RESULTS.value

    def archived_date(self, operator: Enum, archived_date: Arrow | datetime | int | str):
        """Filter Archived Date based on **archivedDate** keyword.

        Args:
            operator: The operator enum for the filter.
            archived_date: The date the result was archived.
        """
        archived_date = self.util.any_to_datetime(archived_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('archivedDate', operator, archived_date, TqlType.STRING)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date and time that the entity was first created.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    @property
    def has_intel_requirement(self):
        """Return **IntelRequirementFilter** for further filtering."""
        # first-party
        from tcex.api.tc.v3.intel_requirements.intel_requirement_filter import (
            IntelRequirementFilter,
        )

        intel_requirements = IntelRequirementFilter(Tql())
        self._tql.add_filter(
            'hasIntelRequirement', TqlOperator.EQ, intel_requirements, TqlType.SUB_QUERY
        )
        return intel_requirements

    def id(self, operator: Enum, id: int | list):  # noqa: A002
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the intel query result.
        """
        if isinstance(id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def intel_id(self, operator: Enum, intel_id: int | list):
        """Filter Intel ID based on **intelId** keyword.

        Args:
            operator: The operator enum for the filter.
            intel_id: The ID of the entity related to the result.
        """
        if isinstance(intel_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('intelId', operator, intel_id, TqlType.INTEGER)

    def intel_req_id(self, operator: Enum, intel_req_id: int | list):
        """Filter ID based on **intelReqId** keyword.

        Args:
            operator: The operator enum for the filter.
            intel_req_id: The ID of the intel requirement.
        """
        if isinstance(intel_req_id, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('intelReqId', operator, intel_req_id, TqlType.INTEGER)

    def intel_type(self, operator: Enum, intel_type: list | str):
        """Filter Intel Type based on **intelType** keyword.

        Args:
            operator: The operator enum for the filter.
            intel_type: The intel type of the result.
        """
        if isinstance(intel_type, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('intelType', operator, intel_type, TqlType.STRING)

    def is_archived(self, operator: Enum, is_archived: bool):
        """Filter Archived Flag based on **isArchived** keyword.

        Args:
            operator: The operator enum for the filter.
            is_archived: A true/false indicating if the result has been archived.
        """
        self._tql.add_filter('isArchived', operator, is_archived, TqlType.BOOLEAN)

    def is_associated(self, operator: Enum, is_associated: bool):
        """Filter Associated based on **isAssociated** keyword.

        Args:
            operator: The operator enum for the filter.
            is_associated: A true/false indicating if the result has been associated.
        """
        self._tql.add_filter('isAssociated', operator, is_associated, TqlType.BOOLEAN)

    def is_deleted(self, operator: Enum, is_deleted: bool):
        """Filter Deleted based on **isDeleted** keyword.

        Args:
            operator: The operator enum for the filter.
            is_deleted: A true/false indicating if the result has been deleted.
        """
        self._tql.add_filter('isDeleted', operator, is_deleted, TqlType.BOOLEAN)

    def is_false_positive(self, operator: Enum, is_false_positive: bool):
        """Filter False Positive based on **isFalsePositive** keyword.

        Args:
            operator: The operator enum for the filter.
            is_false_positive: A true/false indicating if the result has been flagged as false
                positive.
        """
        self._tql.add_filter('isFalsePositive', operator, is_false_positive, TqlType.BOOLEAN)

    def is_local(self, operator: Enum, is_local: bool):
        """Filter Sourced Internally based on **isLocal** keyword.

        Args:
            operator: The operator enum for the filter.
            is_local: A true/false indicating if the result came from ThreatConnect repository.
        """
        self._tql.add_filter('isLocal', operator, is_local, TqlType.BOOLEAN)

    def last_matched_date(self, operator: Enum, last_matched_date: Arrow | datetime | int | str):
        """Filter Last Matched Date based on **lastMatchedDate** keyword.

        Args:
            operator: The operator enum for the filter.
            last_matched_date: The date the result last matched the keyword query.
        """
        last_matched_date = self.util.any_to_datetime(last_matched_date).strftime(
            '%Y-%m-%d %H:%M:%S'
        )
        self._tql.add_filter('lastMatchedDate', operator, last_matched_date, TqlType.STRING)

    def last_modified(self, operator: Enum, last_modified: Arrow | datetime | int | str):
        """Filter Last Modified based on **lastModified** keyword.

        Args:
            operator: The operator enum for the filter.
            last_modified: The date and time the entity was modified.
        """
        last_modified = self.util.any_to_datetime(last_modified).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('lastModified', operator, last_modified, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The Owner ID for the result.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name for the result.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def score(self, operator: Enum, score: float | list):
        """Filter Score based on **score** keyword.

        Args:
            operator: The operator enum for the filter.
            score: The weighted score in the relevancy of the result.
        """
        if isinstance(score, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('score', operator, score, TqlType.FLOAT)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Name based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: The summary of the result.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            ex_msg = (
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )
            raise RuntimeError(ex_msg)

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)



"""TcEx Framework Module"""

# standard library
import collections
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import datetime
from inspect import signature
from typing import Any, cast

# third-party
import jmespath

# first-party
from tcex.api.tc.ti_transform import ti_predefined_functions
from tcex.api.tc.ti_transform.model import (
    AttributeTransformModel,
    GroupTransformModel,
    IndicatorTransformModel,
    MetadataTransformModel,
    SecurityLabelTransformModel,
    TagTransformModel,
)
from tcex.api.tc.ti_transform.model.transform_model import (
    AssociatedGroupTransform,
    DatetimeTransformModel,
    FileOccurrenceTransformModel,
    PredefinedFunctionModel,
)
from tcex.logger.trace_logger import TraceLogger
from tcex.pleb.jmespath_custom import TcFunctions
from tcex.util import Util

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class TransformException(Exception):
    """Base exception for transform errors."""

    def __init__(self, field: str, cause: Exception, context: dict | None, *args) -> None:
        """."""
        super().__init__(*args)
        self.field = field
        self.cause = cause
        self.context = context

    def __str__(self) -> str:
        """."""
        return f'Error transforming {self.field}: {self.cause}'


class NoValidTransformException(Exception):
    """Exception for when no valid transform is found."""


class TransformsABC(ABC):  # noqa: B024
    """Transform Abstract Base Class"""

    def __init__(
        self,
        ti_dicts: list[dict],
        transforms: list[GroupTransformModel | IndicatorTransformModel],
        raise_exceptions: bool = False,
    ):
        """Initialize instance properties."""
        self.ti_dicts = ti_dicts
        self.transforms = transforms

        # properties
        self.log = _logger
        self.raise_exceptions = raise_exceptions
        self.transformed_collection: list[TransformABC] = []

        # validate transforms
        self._validate_transforms()

    def _validate_transforms(self):
        """Validate the transform model."""
        if len(self.transforms) > 1:
            for transform in self.transforms:
                if not callable(transform.applies):
                    ex_msg = (
                        'If more than one transform is provided, each '
                        'provided transform must provide an apply field.'
                    )
                    raise ValueError(ex_msg)  # noqa: TRY004


class TransformABC(ABC):
    """Transform Abstract Base Class"""

    def __init__(
        self,
        ti_dict: dict,
        transforms: list[GroupTransformModel | IndicatorTransformModel],
    ):
        """Initialize instance properties."""
        self.ti_dict = ti_dict
        self.transforms = transforms if isinstance(transforms, list) else [transforms]

        # properties
        self.adhoc_groups: list[dict] = []
        self.adhoc_indicators: list[dict] = []
        self.log = _logger
        # the current active transform
        self.transform: GroupTransformModel | IndicatorTransformModel
        self.transformed_item = {}
        self.util = Util()
        self.jmespath_options = jmespath.Options(
            custom_functions=TcFunctions(), dict_cls=collections.OrderedDict
        )

        # validate transforms
        self._validate_transforms()

    @staticmethod
    def _always_array(value: str | list | None) -> list:
        """Ensure value is always an array."""
        if value is None:
            value = []
        elif not isinstance(value, list):
            value = [value]
        return value

    @staticmethod
    def _build_summary(
        val1: str | None = None, val2: str | None = None, val3: str | None = None
    ) -> str:
        """Build the Indicator summary using available values."""
        return ' : '.join([value for value in [val1, val2, val3] if value is not None])

    def _path_search(self, path: str) -> Any:
        """Return the value of the provided path.

        Path can return any type of data from the TI dict.
        """
        if path is not None:
            return jmespath.search(path, self.ti_dict, options=self.jmespath_options)
            # self.log.trace(f'feature=transform, action=path-search, path={path}, value={value}')
        return None

    def _process(self):
        """Process the TI data."""
        # choose the correct transform model before processing TI data
        self._select_transform()

        # process type first, fail early
        self._process_type()

        # process type specific data
        if isinstance(self.transform, GroupTransformModel):
            self._process_group()
        elif isinstance(self.transform, IndicatorTransformModel):
            self._process_indicator()

        # self.process_associations(self.transform.associations)
        self._process_associated_group(self.transform.associated_groups)
        self._process_attributes(self.transform.attributes or [])
        self._process_security_labels(self.transform.security_labels or [])
        self._process_tags(self.transform.tags or [])

        # date fields
        # date fields
        self._process_metadata_datetime('dateAdded', self.transform.date_added)
        self._process_metadata_datetime('lastModified', self.transform.last_modified)
        self._process_metadata_datetime('firstSeen', self.transform.first_seen)
        self._process_metadata_datetime('lastSeen', self.transform.first_seen)
        self._process_metadata_datetime('externalDateAdded', self.transform.external_date_added)
        self._process_metadata_datetime('externalDateExpires', self.transform.external_date_expires)
        self._process_metadata_datetime(
            'externalLastModified', self.transform.external_last_modified
        )
        self._process_metadata_datetime('firstSeen', self.transform.first_seen)
        self._process_metadata_datetime('lastSeen', self.transform.first_seen)
        self._process_metadata_datetime('externalDateAdded', self.transform.external_date_added)
        self._process_metadata_datetime('externalDateExpires', self.transform.external_date_expires)
        self._process_metadata_datetime(
            'externalLastModified', self.transform.external_last_modified
        )

        # xid
        self._process_metadata('xid', self.transform.xid)

    def _process_associated_group(self, associations: list[AssociatedGroupTransform]):
        """Process Attribute data"""
        for i, association in enumerate(associations or [], 1):
            try:
                for value in filter(
                    bool, self._process_metadata_transform_model(association.value)
                ):
                    self.add_associated_group(value)  # type: ignore
            except Exception as ex:
                ex_msg = f'Associated Group [{i}]'
                raise TransformException(ex_msg, ex, context=association.dict()) from ex

    def _process_metadata_transform_model(
        self, value: bool | MetadataTransformModel | str | None, expected_length: int | None = None
    ) -> list:
        """Process fields that can be static values or a MetadataTransformModel.

        If value is not a MetadataTransformModel (i.e., it's a static value), and expected_length
        is given, "spread" the static value into an array of expected_length length.

        """
        if value is None:
            if expected_length:
                return [value] * expected_length
            return []

        if isinstance(value, MetadataTransformModel):
            transformed_value = self._transform_values(value)

            if expected_length is not None and len(transformed_value) != expected_length:
                ex_msg = (
                    f'Expected transform value of length {expected_length} for {value}, '
                    f'but length was {len(transformed_value)}'
                )
                raise RuntimeError(ex_msg)

            return transformed_value

        return [value] * expected_length if expected_length is not None else [value]

    def _process_attributes(self, attributes: list[AttributeTransformModel]):
        """Process Attribute data"""
        for i, attribute in enumerate(attributes or [], start=1):
            try:
                values = self._process_metadata_transform_model(attribute.value)
                if not values:
                    # self.log.info(f'No values found for attribute transform {attribute.dict()}')
                    continue

                types = self._process_metadata_transform_model(attribute.type, len(values))
                source = self._process_metadata_transform_model(attribute.source, len(values))
                displayed = self._process_metadata_transform_model(attribute.displayed, len(values))
                pinned = self._process_metadata_transform_model(attribute.pinned, len(values))

                param_keys = ['type_', 'value', 'displayed', 'pinned', 'source']
                params = [
                    dict(zip(param_keys, p, strict=False))
                    for p in zip(types, values, displayed, pinned, source, strict=False)
                ]

                for param in params:
                    param_ = self.util.remove_none(param)
                    if 'value' not in param_:
                        continue

                    if 'type_' not in param_:
                        self.log.warning(
                            'feature=transform, action=process-attribute, '
                            f'transform={attribute.dict(exclude_unset=True)}, error=no-type'
                        )
                        continue

                    # strip out None params so that required params are enforced and optional
                    # params with default values are respected.
                    try:
                        self.add_attribute(**param_)
                    except Exception:
                        self.log.exception(
                            'feature=transform, action=process-attribute, '
                            f'transform={attribute.dict(exclude_unset=True)}'
                        )
            except Exception as ex:
                ex_msg = f'Attribute [{i}], type={attribute.type}'
                raise TransformException(ex_msg, ex, context=attribute.dict()) from ex

    def _process_file_occurrences(self, file_occurrences: list[FileOccurrenceTransformModel]):
        """Process File Occurrences data.

        File Occurrences are a bit weird, in that none of the fields are required.  Because of this,
        There may be results where all of the fields are None, in which case we'll skip that result
        and not call add_file_occurrence.
        """
        for file_occurrence in file_occurrences or []:
            expected_length = max(
                len(self._process_metadata_transform_model(f))
                for f in [
                    field
                    for field in (
                        file_occurrence.file_name,
                        file_occurrence.path,
                        file_occurrence.date,
                    )
                    if field is not None
                ]
            )
            file_name = self._process_metadata_transform_model(
                file_occurrence.file_name, expected_length
            )
            path = self._process_metadata_transform_model(file_occurrence.path, expected_length)
            date = self._process_metadata_transform_model(file_occurrence.date, expected_length)

            param_keys = ['file_name', 'path', 'date']
            params = [
                dict(zip(param_keys, p, strict=False))
                for p in zip(file_name, path, date, strict=False)
            ]

            for kwargs in filter(bool, params):  # get rid of empty dicts
                self.add_file_occurrence(**self.util.remove_none(kwargs))

    def _process_confidence(self, metadata: MetadataTransformModel | None):
        """Process standard metadata fields."""
        self.add_confidence(self._transform_value(metadata))

    def _process_group(self):
        """Process Group Specific data."""
        if not isinstance(self.transform, GroupTransformModel):
            return

        self._process_name()

        if self.transformed_item['type'] == 'Campaign':
            self._process_metadata_datetime('firstSeen', self.transform.first_seen)

        if self.transformed_item['type'] == 'Document':
            self._process_metadata('fileName', self.transform.file_name)
            self._process_metadata('malware', self.transform.malware)
            self._process_metadata('password', self.transform.password)

        if self.transformed_item['type'] == 'Email':
            self._process_metadata('body', self.transform.body)
            self._process_metadata('from', self.transform.from_addr)
            self._process_metadata('header', self.transform.header)
            self._process_metadata('subject', self.transform.subject)
            self._process_metadata('to', self.transform.to_addr)

        if self.transformed_item['type'] in ('Event', 'Incident'):
            self._process_metadata_datetime('eventDate', self.transform.event_date)
            self._process_metadata('status', self.transform.status)

        if self.transformed_item['type'] == 'Report':
            self._process_metadata('fileName', self.transform.file_name)
            self._process_metadata_datetime('publishDate', self.transform.publish_date)

        # Handle sig specific fields here
        if self.transformed_item['type'] == 'Signature':
            self._process_metadata('fileName', self.transform.file_name)
            self._process_metadata('fileType', self.transform.file_type)
            self._process_metadata('fileText', self.transform.file_text)

    def _process_indicator(self):
        """Process Indicator Specific data."""
        if not isinstance(self.transform, IndicatorTransformModel):
            return

        # handle the 3 possible indicator fields
        self._process_indicator_values()

        if self.transform.active:
            self._process_metadata('active', self.transform.active)

        self._process_confidence(self.transform.confidence)
        self._process_rating(self.transform.rating)

        if self.transformed_item['type'] == 'File':
            self._process_metadata('size', self.transform.size)
            self._process_file_occurrences(self.transform.file_occurrences or [])

        if self.transformed_item['type'] == 'Host':
            self._process_metadata('dnsActive', self.transform.dns_active)
            self._process_metadata('whoisActive', self.transform.whois_active)

    def _process_indicator_values(self):
        """Process Indicator value."""
        if not isinstance(self.transform, IndicatorTransformModel):
            return

        value1 = self._transform_value(self.transform.value1)
        value2 = self._transform_value(self.transform.value2)
        value3 = self._transform_value(self.transform.value3)

        if not any([value1, value2, value3]):
            self.log.error(
                'feature=ti-transform, event=process-indicators, message=no-indicator-value-found, '
                f'path-value1={value1}, path-value2={value2}, path-value3={value3}'
            )
            ex_msg = 'At least one indicator value must be provided.'
            raise RuntimeError(ex_msg)

        self.add_summary(self._build_summary(value1, value2, value3))

    def _process_name(self):
        """Process Group Name data."""
        if not isinstance(self.transform, GroupTransformModel):
            return

        name = self._transform_value(self.transform.name)

        if name is None:
            self.log.error(
                'feature=ti-transform, event=process-group-name, message=no-name-found, '
                f'path={self.transform.name.path}'
            )
            ex_msg = 'At least one indicator value must be provided.'
            raise RuntimeError(ex_msg)

        self.add_name(name)

    def _process_metadata(self, key: str, metadata: MetadataTransformModel | None):
        """Process standard metadata fields."""
        try:
            value = self._transform_value(metadata)
            if value is not None:
                self.add_metadata(key, value)
        except Exception as ex:
            raise TransformException(key, ex, metadata.dict() if metadata else None) from ex

    def _process_metadata_datetime(self, key: str, metadata: DatetimeTransformModel | None):
        """Process metadata fields that should be a TC datetime."""
        try:
            if metadata is not None and metadata.path is not None:
                value = self._path_search(metadata.path)
                if value is not None:
                    self.add_metadata(
                        key, self.util.any_to_datetime(value).strftime('%Y-%m-%dT%H:%M:%SZ')
                    )
        except Exception as ex:
            raise TransformException(key, ex, context=metadata.dict() if metadata else None) from ex

    def _process_security_labels(self, labels: list[SecurityLabelTransformModel]):
        """Process Tag data"""
        for i, label in enumerate(labels or [], 1):
            try:
                names = self._process_metadata_transform_model(label.value)
                if not names:
                    # self.log.info(f'No values found for security label transform {label.dict()}')
                    continue

                descriptions = self._process_metadata_transform_model(
                    label.description, expected_length=len(names)
                )
                colors = self._process_metadata_transform_model(
                    label.color, expected_length=len(names)
                )

                param_keys = ['color', 'description', 'name']
                params = [
                    dict(zip(param_keys, p, strict=False))
                    for p in zip(colors, descriptions, names, strict=False)
                ]

                for kwargs in params:
                    kwargs_ = self.util.remove_none(kwargs)
                    if 'name' not in kwargs_:
                        continue
                    # strip out None params so that required params are enforced and optional
                    # params with default values are respected.
                    self.add_security_label(**kwargs_)
            except Exception as ex:
                ex_msg = f'Security Labels [{i}]'
                raise TransformException(ex_msg, ex, context=label.dict()) from ex

    def _process_tags(self, tags: list[TagTransformModel]):
        """Process Tag data"""
        for i, tag in enumerate(tags or [], 1):
            try:
                for value in filter(bool, self._process_metadata_transform_model(tag.value)):
                    self.add_tag(name=value)  # type: ignore
            except Exception as ex:
                ex_msg = f'Tags [{i}]'
                raise TransformException(ex_msg, ex, context=tag.dict()) from ex

    def _process_rating(self, metadata: MetadataTransformModel | None):
        """Process standard metadata fields."""
        try:
            self.add_rating(self._transform_value(metadata))
        except Exception as ex:
            ex_msg = 'Rating'
            raise TransformException(
                ex_msg, ex, context=metadata.dict() if metadata else None
            ) from ex

    def _process_type(self):
        """Process standard metadata fields."""
        try:
            _type = self._transform_value(self.transform.type)
            if _type is not None:
                self.transformed_item['type'] = _type
            else:
                self.log.error(
                    'feature=ti-transform, action=process-type, error=invalid=type, '
                    f'path={self.transform.type.path}, value={_type}'
                )
                ex_msg = 'Invalid type'
                raise RuntimeError(ex_msg)  # noqa: TRY301
        except Exception as ex:
            ex_msg = 'Type'
            raise TransformException(ex_msg, ex, context=self.transform.type.dict()) from ex

    def _select_transform(self):
        """Select the correct transform based on the "applies" field."""
        for transform in self.transforms:
            if transform.applies is None or transform.applies(self.ti_dict) is True:
                self.transform = transform
                break
        else:
            ex_msg = 'No transform found for TI data'
            raise NoValidTransformException(ex_msg)

    def _transform_value(self, metadata: MetadataTransformModel | None) -> str | None:
        """Pass value to series transforms."""
        # not all fields are required
        if metadata is None:
            return None

        # not all metadata fields have a path, but they must have a path or default
        if metadata.path is None:
            return metadata.default

        # get value from path
        value = self._path_search(metadata.path)

        # return default if value of None is returned from Path
        # IMPORTANT: a value of None passed to the transform may cause a failure (lambda x.lower())
        if value is None:
            return metadata.default

        for t in metadata.transform or []:
            # pass value to static_map or callable, but never both
            value = cast(Any, value)  # due to line 472, we know value is not None.
            if t.filter_map is not None:
                value = self._transform_value_map(value, t.filter_map, passthrough=True)
            elif t.static_map is not None:
                value = self._transform_value_map(value, t.static_map)
            elif callable(t.method):
                value = self._transform_value_callable(value, t.method, t.kwargs)

        # ensure only a string value or None is returned (set to default if required)
        if value is None:
            value = metadata.default
        elif not isinstance(value, str):
            value = str(value)

        return value

    def _transform_value_callable(
        self, value: dict | list | str, c: Callable | PredefinedFunctionModel, kwargs=None
    ) -> str | None | list[str]:
        """Transform values in the TI data."""
        # find signature of method and call with correct args
        if isinstance(c, PredefinedFunctionModel):
            normalized_params = {
                k.replace(' ', '_').lower(): v for k, v in PredefinedFunctionModel.params or {}
            }
            return getattr(ti_predefined_functions, c.name)(value, **normalized_params)

        kwargs = kwargs or {}
        try:
            sig = signature(c, follow_wrapped=True)
            if 'ti_dict' in sig.parameters:
                kwargs.update({'ti_dict': self.ti_dict})
            if 'transform' in sig.parameters:
                kwargs.update({'transform': self})
        except ValueError:  # signature doesn't work for many built-in methods/functions
            pass

        # pass value to transform callable/method, which should always return a string
        return c(value, **kwargs)

    def _transform_value_map(self, value: str, map_: dict, passthrough: bool = False) -> str:
        """Transform a value using a static map."""
        _default = value if passthrough is True else None
        if isinstance(value, str):
            # a static map is a dict of key/value pairs
            value = map_.get(value.lower(), _default)
        else:
            self.log.warning(
                f"""feature=ti-transform, action=transform-value, """
                f"""message='static-map-requires-str-value", value={value}"""
            )
        return value

    def _transform_values(self, metadata: MetadataTransformModel | None) -> list[str]:
        """Pass value to series transforms."""

        def _default() -> list:
            """Return default value (as list) if exists, else empty list."""
            if metadata is None or metadata.default is None:
                return []

            return self._always_array(metadata.default)

        # not all items have all metadata fields
        if metadata is None:
            # self.log.trace('feature=transform, action=transform-values, metadata=None')
            return []

        # not all metadata fields have a path, but they must have a path or default
        if metadata.path is None:
            return _default()
            # self.log.trace(
            #     f'feature=transform, action=transform-values, metadata-path=None, value={default}'
            # )

        # path search can return multiple data types and single or multiple values
        value = self._path_search(metadata.path)

        # return default if value of None is returned from Path
        # IMPORTANT: a None value passed to the transform may cause a failure (lambda x: x.lower())
        if value in [None, [], '']:
            return _default()
            # self.log.trace(
            #     f'feature=transform, action=transform-values, metadata-path=None, value={default}'
            # )

        for t in metadata.transform or []:
            if t.filter_map is not None:
                # when path search returns an array of values, each value is mapped
                value = [
                    self._transform_value_map(v, t.filter_map, passthrough=True)
                    for v in self._always_array(value)
                ]
            elif t.static_map is not None:
                # when path search returns an array of values, each value is mapped
                _values = []
                for v in self._always_array(value):
                    v_ = self._transform_value_map(v, t.static_map)
                    if v_ is not None:
                        _values.append(v_)
                value = _values
            # PYRIGHT-MISS - None check for value already performed above
            elif callable(t.method) and value is not None:
                value = self._transform_value_callable(value, t.method, t.kwargs)
            elif callable(t.for_each):
                value = [
                    self._transform_value_callable(v, t.for_each, t.kwargs) if v is not None else v
                    for v in self._always_array(value)
                ]

        # the output should be an array of strings or empty array
        _value = []
        for v in self._always_array(value):
            if v in [None, '']:
                if metadata.default is not None:
                    _value.append(metadata.default)
                else:
                    _value.append(v)
            else:
                _value.append(v)

        # self.log.trace(f'feature=transform, action=transform-values, value={_value}')
        return _value

    def _validate_transforms(self):
        """Validate the transform model."""
        if len(self.transforms) > 1:
            for transform in self.transforms:
                if transform.applies is None:
                    ex_msg = (
                        'If more than one transform is provided, each '
                        'provided transform must provide an apply field.'
                    )
                    raise ValueError(ex_msg)

    @abstractmethod
    def add_associated_group(self, group_xid: str):
        """Abstract method"""

    @abstractmethod
    def add_attribute(
        self,
        type_: str,
        value: str,
        displayed: bool = False,
        pinned: bool = False,
        source: str | None = None,
    ):
        """Abstract method"""

    @abstractmethod
    def add_file_occurrence(
        self,
        file_name: str | None = None,
        path: str | None = None,
        date: datetime | None = None,
    ):
        """Abstract method"""

    @abstractmethod
    def add_confidence(self, confidence: int | str | None):
        """Abstract method"""

    @abstractmethod
    def add_metadata(self, key: str, value: str):
        """Abstract method"""

    @abstractmethod
    def add_name(self, name: str | None):
        """Abstract method"""

    @abstractmethod
    def add_rating(self, rating: float | str | None):
        """Abstract method"""

    @abstractmethod
    def add_security_label(
        self, name: str, color: str | None = None, description: str | None = None
    ):
        """Abstract method"""

    @abstractmethod
    def add_summary(self, value: str | None):
        """Abstract method"""

    @abstractmethod
    def add_tag(self, name: str):
        """Abstract method"""



"""TcEx Framework Module"""

# standard library
import importlib
from abc import ABC
from typing import Any

# first-party
from tcex.api.tc.v3._gen._gen_abc import GenerateABC
from tcex.util.render.render import Render


class GenerateArgsABC(GenerateABC, ABC):
    """Generate docstring for Model."""

    def __init__(self, type_: Any):
        """Initialize instance properties."""
        super().__init__(type_)

    @staticmethod
    def _import_model(module: Any, class_name: str) -> Any:
        """Import the appropriate model."""
        return getattr(importlib.import_module(module), class_name)

    def _prop_type(self, prop_data: dict[str, dict | str | None]) -> str | None:
        """Return the appropriate arg type."""
        prop_type = None
        if 'type' in prop_data and prop_data.get('type') is not None:
            prop_type = self._prop_type_map(prop_data.get('type'))  # type: ignore
        elif 'allOf' in prop_data and prop_data.get('allOf'):
            ref: str = prop_data.get('allOf')[0].get('$ref')  # type: ignore
            prop_type = ref.split('/')[-1].replace('Model', '')
        elif 'items' in prop_data and prop_data.get('items'):
            ref: str = prop_data.get('items', {}).get('$ref')  # type: ignore
            prop_type = ref.split('/')[-1].replace('Model', '')
        return prop_type

    @staticmethod
    def _prop_type_map(prop_type: str) -> str:
        """Return hint type."""
        _prop_types = {
            'boolean': 'bool',
            'integer': 'int',
            'string': 'str',
        }
        return _prop_types.get(prop_type, prop_type)

    def gen_args(
        self,
        i1: str | None = None,
        i2: str | None = None,
        updatable: bool = True,
    ) -> str:
        """Model Map"""
        i1 = i1 or self.i1
        i2 = i2 or self.i2

        module_import_data = self._module_import_data(self.type_)
        model = self._import_model(
            module_import_data['model_module'], module_import_data['model_class']
        )
        _doc_string = [f'{i1}Args:']

        # get properties from schema
        schema = model.schema(by_alias=False)
        properties = {}
        if '$ref' in schema:
            model_name = schema.get('$ref').split('/')[-1]
            properties = schema.get('definitions').get(model_name).get('properties')
        elif 'properties' in schema:
            properties = schema.get('properties')
        else:
            Render.panel.failure(model.schema_json(by_alias=False))

        # iterate over properties to build docstring
        for arg, prop_data in properties.items():
            # for all doc string read-only args should not be included.
            if prop_data.get('read_only', False) is True:
                continue

            # for add_xxx method doc string non-updatable args should not be included.
            if updatable is False and prop_data.get('updatable', True) is False:
                continue

            # get arg type
            prop_type = self._prop_type(prop_data)

            # arg
            _arg_doc = f'{arg} ({prop_type}, kwargs)'

            # description
            description = prop_data.get('description')
            _arg_doc = self._format_description(
                arg=_arg_doc,
                description=description,
                length=100,
                indent=' ' * len(i2),
            )

            # add arg to doc string
            _doc_string.append(f'{i2}{_arg_doc}')

        if len(_doc_string) > 1:
            return '\n'.join(_doc_string)
        return ''



"""TcEx Framework Module"""

# standard library
import concurrent.futures
import json
import sys
import threading
from collections.abc import Callable
from functools import reduce
from io import BytesIO
from typing import Any

# first-party
from tcex.app.key_value_store.key_value_store import KeyValueStore
from tcex.app.service.common_service import CommonService
from tcex.app.token import Token
from tcex.input.field_type.sensitive import Sensitive
from tcex.input.model.module_app_model import ModuleAppModel
from tcex.logger.logger import Logger


class ApiService(CommonService):
    """TcEx Framework API Service module."""

    def __init__(
        self,
        key_value_store: KeyValueStore,
        logger: Logger,
        model: ModuleAppModel,
        token: Token,
    ):
        """Initialize the Class properties."""
        super().__init__(key_value_store, logger, model, token)

        # properties
        self._metrics = {'Errors': 0, 'Requests': 0, 'Responses': 0}

        # config callbacks
        self.api_event_callback: Any = None

        # thread pool to handle HTTP requests
        self.request_thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=5)

    @property
    def command_map(self) -> dict:
        """Return the command map for the current Service type."""
        command_map = super().command_map
        command_map.update({'runservice': self.process_run_service_command})
        return command_map

    def format_query_string(self, params: dict) -> str:
        """Convert name/value array to a query string.

        Args:
            params: The query params for the request.

        Returns:
            str: The query params reformatted as a string.
        """
        query_string = []
        try:
            for q in params:
                query_string.append(f"""{q.get('name')}={q.get('value')}""")
        except AttributeError:
            self.log.exception(f'feature=api-service, event=bad-params-provided, params={params}')
        return '&'.join(query_string)

    def format_request_headers(self, headers: dict) -> dict:
        """Convert name/value array to a headers dict.

        Args:
            headers: The dict of key/value header data.

        Returns:
            dict: The restructured header data.
        """
        headers_ = {}
        try:
            for h in headers:
                # TODO: either support tuple or csv list of values
                # headers_.setdefault(h.get('name').lower(), []).append(h.get('value'))
                headers_.setdefault(h.get('name').lower(), str(h.get('value')))

        except AttributeError:
            self.log.exception(
                f'feature=api-service, event=bad-headers-provided, headers={headers}'
            )
        return headers_

    def format_response_headers(self, headers: dict) -> list[dict]:
        """Convert name/value array to a query string.

        Args:
            headers: The dict header data to be converted to key/value pairs.

        Returns:
            dict: The restructured header data.
        """
        headers_ = []
        try:
            for h in headers:
                headers_.append({'name': h[0], 'value': h[1]})
        except AttributeError:
            self.log.exception(
                f'feature=api-service, event=bad-headers-provided, headers={headers}'
            )
        return headers_

    def process_run_service_response(self, *args, **kwargs):
        """Handle service event responses.

        ('200 OK', [('content-type', 'application/json'), ('content-length', '103')])
        """
        self.log.info('feature=api-service, event=response-received, status=waiting-for-body')
        kwargs['event'].wait(30)  # wait for thread event - (set on body write)
        self.log.trace(f'feature=api-service, event=response, args={args}')
        try:
            status_code, status = args[0].split(' ', 1)
            response = {
                'bodyVariable': 'response.body',
                'command': 'Acknowledged',
                'headers': self.format_response_headers(args[1]),
                'requestKey': kwargs.get('request_key'),
                'status': status,
                'statusCode': status_code,
                'type': 'RunService',
            }
            self.log.info('feature=api-service, event=response-sent')
            self.message_broker.publish(json.dumps(response), self.model.tc_svc_client_topic)
            self.increment_metric('Responses')
        except Exception:
            self.log.exception('feature=api-service, event=failed-creating-response-body')
            self.increment_metric('Errors')

    def process_run_service_command(self, message: dict):
        """Process the RunService command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
              "command": "RunService",
              "apiToken": "abc123",
              "bodyVariable": "request.body",
              "headers": [ { key/value pairs } ],
              "method": "GET",
              "queryParams": [ { key/value pairs } ],
              "requestKey": "123abc",
              "userConfig": [{
                "name": "tlpExportSetting",
                "value": "TLP:RED"
              }],
            }

        Args:
            message: The message payload from the server topic.
        """
        # register config apiToken (before any logging)
        self.token.register_token(
            self.thread_name, Sensitive(message['apiToken']), message.get('expireSeconds')
        )
        self.log.info(f'feature=api-service, event=runservice-command, message="{message}"')

        # thread event used to block response until body is written
        event = threading.Event()

        # process message
        request_key: str = message['requestKey']
        body = None
        try:
            # read body from redis
            body_variable: str = message.pop('bodyVariable', None)
            if body_variable is not None:
                body: Any = self.key_value_store.client.read(request_key, body_variable)
                if body is not None:
                    # for API service the data in Redis is not b64 encoded
                    body = BytesIO(body)
        except Exception:
            self.log.exception('feature=api-service, event=failed-reading-body')
        headers = self.format_request_headers(message.pop('headers'))
        method: str = message.pop('method')
        params: dict = message.pop('queryParams')
        path: str = message.pop('path')

        try:
            environ = {
                'wsgi.errors': sys.stderr,
                'wsgi.input': body,
                'wsgi.multithread': True,
                'wsgi.multiprocess': False,
                'wsgi.run_once': True,
                'wsgi.url_scheme': 'https',
                'wsgi.version': (1, 0),
                'PATH_INFO': path,
                'QUERY_STRING': self.format_query_string(params),
                'REMOTE_ADDR': message.get('remoteAddress', ''),
                # 'REMOTE_HOST': message.get('remoteAddress', ''),
                'REQUEST_METHOD': method.upper(),
                'SCRIPT_NAME': '/',
                'SERVER_NAME': '',
                'SERVER_PORT': '',
                'SERVER_PROTOCOL': 'HTTP/1.1',
            }

            # Add user config for TAXII or other service that supports the data type
            environ['user_config'] = message.get('userConfig', [])

            # add headers
            if headers.get('content-type') is not None:
                environ['CONTENT_TYPE'] = headers.pop('content-type')

            # add content length
            if headers.get('content-length') is not None:
                environ['CONTENT_LENGTH'] = headers.pop('content-length')

            for header, value in headers.items():
                environ[f'HTTP_{header}'.upper()] = value

            # make values from message available in env in camel
            # case (e.g., falcon -> req.env.get('request_url))
            for key, value in message.items():
                if key not in environ and self.util.camel_to_snake(key) not in environ:
                    environ[self.util.camel_to_snake(key)] = value

            self.log.trace(f'feature=api-service, environ={environ}')
            self.increment_metric('Requests')
        except Exception:
            self.log.exception('feature=api-service, event=failed-building-environ')
            self.increment_metric('Errors')
            return  # stop processing

        def response_handler(*args, **kwargs):
            """Handle WSGI Response"""
            kwargs['event'] = event  # add event to kwargs for blocking
            kwargs['request_key'] = request_key
            self.service_thread(
                name='response-handler',
                target=self.process_run_service_response,
                args=args,
                kwargs=kwargs,
            )

        if callable(self.api_event_callback):
            try:
                body_data: Any = self.api_event_callback(environ, response_handler)
                if body_data:
                    body_data = reduce(lambda a, b: a + b, body_data)

                # write body to Redis
                if body_data:
                    self.key_value_store.client.create(request_key, 'response.body', body_data)

                    # set thread event to True to trigger response
                    self.log.info('feature=api-service, event=response-body-written')

                # release event lock
                event.set()
            except Exception:
                self.log.exception('feature=api-service, event=api-event-callback-failed')
                self.increment_metric('Errors')

        # unregister config apiToken
        self.token.unregister_token(self.thread_name)

    def process_shutdown_command(self, message: dict):
        """Handle shutdown command."""
        super().process_shutdown_command(message)
        self.request_thread_pool.shutdown()

    def service_thread(
        self,
        name: str,
        target: Callable[..., bool | None],
        args: tuple | None = None,
        kwargs: dict | None = None,
        session_id: str | None = None,
        trigger_id: int | None = None,
    ):
        """If this is a run-service command, run it with the thread pool.

        For everything else, use parent's implementation.

        Args:
            name: The name of the thread.
            target: The method to call for the thread.
            args: The args to pass to the target method.
            kwargs: Additional args.
            session_id: The current session id.
            trigger_id: The current trigger id.
        """
        args = args or ()
        kwargs = kwargs or {}

        def _thread_wrapper():
            t = threading.current_thread()
            t.name = name
            target(*args, **kwargs)

        if target is self.process_run_service_command:
            self.log.info(f'feature=service, event=request-thread-creation, name={name}')
            self.request_thread_pool.submit(_thread_wrapper)
        else:
            super().service_thread(name, target, args, kwargs, session_id, trigger_id)



"""TcEx Framework Module"""

# standard library
from collections.abc import Generator
from typing import Any

# third-party
from pydantic.fields import ModelField  # TYPE-CHECKING
from pydantic.types import OptionalInt

# first-party
from tcex.input.field_type.exception import InvalidIntegerValue, InvalidType


class Integer(int):
    """Integer Field Type"""

    ge: OptionalInt = None
    gt: OptionalInt = None
    le: OptionalInt = None
    lt: OptionalInt = None

    @classmethod
    def __modify_schema__(cls, field_schema: dict[str, Any]):
        """Modify the field schema."""

        def update_not_none(mapping: dict[Any, Any], **update: Any):
            mapping.update({k: v for k, v in update.items() if v is not None})

        update_not_none(
            field_schema,
            exclusiveMinimum=cls.gt,
            exclusiveMaximum=cls.lt,
            minimum=cls.ge,
            maximum=cls.le,
        )

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Run validators / modifiers on input."""
        yield cls.validate_type
        yield cls.validate_value

    @classmethod
    def validate_type(cls, value: int | str, field: ModelField) -> int | str:
        """Raise exception if value is not a String type."""
        if not isinstance(value, int | str):
            raise InvalidType(
                field_name=field.name, expected_types='(int, str)', provided_type=type(value)
            )
        return value

    @classmethod
    def validate_value(cls, value: int | str, field: ModelField) -> int:
        """Raise exception if value does not meet criteria."""
        if isinstance(value, str):
            value = int(value)

        if cls.ge is not None and not value >= cls.ge:
            raise InvalidIntegerValue(
                field_name=field.name, operation='greater than or equal to', constraint=cls.ge
            )
        if cls.gt is not None and not value > cls.gt:
            raise InvalidIntegerValue(
                field_name=field.name, operation='greater than', constraint=cls.gt
            )

        if cls.le is not None and not value <= cls.le:
            raise InvalidIntegerValue(
                field_name=field.name, operation='less than or equal to', constraint=cls.le
            )
        if cls.lt is not None and not value < cls.lt:
            raise InvalidIntegerValue(
                field_name=field.name, operation='less than', constraint=cls.lt
            )
        return value


def integer(
    gt: OptionalInt = None,
    ge: OptionalInt = None,
    lt: OptionalInt = None,
    le: OptionalInt = None,
) -> type[Integer]:
    """Return configured instance of String."""
    namespace = {
        'gt': gt,
        'ge': ge,
        'lt': lt,
        'le': le,
    }
    return type('ConstrainedInteger', (Integer,), namespace)



"""TcEx Framework Module"""

# standard library
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING, Self

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactModel
from tcex.api.tc.v3.case_attributes.case_attribute_model import CaseAttributeModel
from tcex.api.tc.v3.cases.case_filter import CaseFilter
from tcex.api.tc.v3.cases.case_model import CaseModel, CasesModel
from tcex.api.tc.v3.groups.group_model import GroupModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorModel
from tcex.api.tc.v3.notes.note_model import NoteModel
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security.assignee_model import AssigneeModel
from tcex.api.tc.v3.security.assignee_user_group_model import AssigneeUserGroupModel
from tcex.api.tc.v3.security.assignee_user_model import AssigneeUserModel
from tcex.api.tc.v3.security.user_groups.user_group_model import UserGroupModel
from tcex.api.tc.v3.security.users.user_model import UserModel
from tcex.api.tc.v3.tags.tag_model import TagModel
from tcex.api.tc.v3.tasks.task_model import TaskModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.artifacts.artifact import Artifact  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.case_attributes.case_attribute import CaseAttribute  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.groups.group import Group  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.indicators.indicator import Indicator  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.notes.note import Note  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.tags.tag import Tag  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.tasks.task import Task  # CIRCULAR-IMPORT


class Case(ObjectABC):
    """Cases Object.

    Args:
        artifacts (Artifacts, kwargs): A list of Artifacts corresponding to the Case.
        assignee (Assignee, kwargs): The user or group Assignee object for the Case.
        associated_cases (Cases, kwargs): A list of Cases associated with this Case.
        associated_groups (Groups, kwargs): A list of Groups associated with this Case.
        associated_indicators (Indicators, kwargs): A list of Indicators associated with this Case.
        attributes (CaseAttributes, kwargs): A list of Attributes corresponding to the Case.
        case_close_time (str, kwargs): The date and time that the Case was closed.
        case_detection_time (str, kwargs): The date and time that ends the user initiated Case
            duration.
        case_occurrence_time (str, kwargs): The date and time that starts the user initiated Case
            duration.
        case_open_time (str, kwargs): The date and time that the Case was first opened.
        description (str, kwargs): The description of the Case.
        name (str, kwargs): The name of the Case.
        notes (Notes, kwargs): A list of Notes corresponding to the Case.
        resolution (str, kwargs): The Case resolution.
        severity (str, kwargs): The Case severity.
        status (str, kwargs): The Case status.
        tags (Tags, kwargs): A list of Tags corresponding to the Case (NOTE: Setting this parameter
            will replace any existing tag(s) with the one(s) specified).
        tasks (Tasks, kwargs): A list of Tasks corresponding to the Case.
        user_access (Users, kwargs): A list of Users that, when defined, are the only ones allowed
            to view or edit the Case.
        workflow_events (WorkflowEvents, kwargs): A list of workflowEvents (timeline) corresponding
            to the Case.
        workflow_template (WorkflowTemplate, kwargs): The Template that the Case is populated by.
        xid (str, kwargs): The **xid** for the Case.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: CaseModel = CaseModel(**kwargs)
        self._nested_field_name = 'cases'
        self._nested_filter = 'has_case'
        self.type_ = 'Case'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.CASES.value

    @property
    def model(self) -> CaseModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | CaseModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            ex_msg = f'Invalid data type: {type(data)} provided.'
            raise RuntimeError(ex_msg)  # noqa: TRY004

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.name}

    @property
    def artifacts(self) -> Generator['Artifact', None, None]:
        """Yield Artifact from Artifacts."""
        # first-party
        from tcex.api.tc.v3.artifacts.artifact import Artifacts

        yield from self._iterate_over_sublist(Artifacts)  # type: ignore

    @property
    def associated_cases(self) -> Generator[Self, None, None]:
        """Yield Case from Cases."""
        # Ensure the current item is not returned as a association
        for case in self._iterate_over_sublist(Cases):  # type: ignore
            if case.model.id == self.model.id:
                continue
            yield case  # type: ignore

    @property
    def associated_groups(self) -> Generator['Group', None, None]:
        """Yield Group from Groups."""
        # first-party
        from tcex.api.tc.v3.groups.group import Groups

        yield from self._iterate_over_sublist(Groups)  # type: ignore

    @property
    def associated_indicators(self) -> Generator['Indicator', None, None]:
        """Yield Indicator from Indicators."""
        # first-party
        from tcex.api.tc.v3.indicators.indicator import Indicators

        yield from self._iterate_over_sublist(Indicators)  # type: ignore

    @property
    def attributes(self) -> Generator['CaseAttribute', None, None]:
        """Yield Attribute from Attributes."""
        # first-party
        from tcex.api.tc.v3.case_attributes.case_attribute import CaseAttributes

        yield from self._iterate_over_sublist(CaseAttributes)  # type: ignore

    @property
    def notes(self) -> Generator['Note', None, None]:
        """Yield Note from Notes."""
        # first-party
        from tcex.api.tc.v3.notes.note import Notes

        yield from self._iterate_over_sublist(Notes)  # type: ignore

    @property
    def tags(self) -> Generator['Tag', None, None]:
        """Yield Tag from Tags."""
        # first-party
        from tcex.api.tc.v3.tags.tag import Tags

        yield from self._iterate_over_sublist(Tags)  # type: ignore

    @property
    def tasks(self) -> Generator['Task', None, None]:
        """Yield Task from Tasks."""
        # first-party
        from tcex.api.tc.v3.tasks.task import Tasks

        yield from self._iterate_over_sublist(Tasks)  # type: ignore

    def stage_artifact(self, data: dict | ObjectABC | ArtifactModel):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = ArtifactModel(**data)

        if not isinstance(data, ArtifactModel):
            ex_msg = 'Invalid type passed in to stage_artifact'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.artifacts.data.append(data)  # type: ignore

    def stage_assignee(
        self,
        type: str,  # noqa: A002
        data: dict | ObjectABC | AssigneeModel | UserModel | UserGroupModel,
    ):
        """Stage artifact on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif type.lower() == 'user' and isinstance(data, dict):
            data = AssigneeModel(type='User', data=data)  # type: ignore
        elif type.lower() == 'group' and isinstance(data, dict):
            data = AssigneeModel(type='Group', data=data)  # type: ignore

        if not isinstance(
            data,
            AssigneeModel | AssigneeUserModel | AssigneeUserGroupModel | UserModel | UserGroupModel,
        ):
            ex_msg = 'Invalid type passed in to stage_assignee'
            raise RuntimeError(ex_msg)  # noqa: TRY004

        if isinstance(data, AssigneeModel):
            self.model.assignee = data
        elif isinstance(
            data, AssigneeUserModel | AssigneeUserGroupModel | UserModel | UserGroupModel
        ):
            self.model.assignee.data = data  # type: ignore
        self.model.assignee.data._staged = True  # type: ignore  # noqa: SLF001

    def stage_associated_case(self, data: dict | ObjectABC | CaseModel):
        """Stage case on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = CaseModel(**data)

        if not isinstance(data, CaseModel):
            ex_msg = 'Invalid type passed in to stage_associated_case'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_cases.data.append(data)  # type: ignore

    def stage_associated_group(self, data: dict | ObjectABC | GroupModel):
        """Stage group on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = GroupModel(**data)

        if not isinstance(data, GroupModel):
            ex_msg = 'Invalid type passed in to stage_associated_group'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_groups.data.append(data)  # type: ignore

    def stage_associated_indicator(self, data: dict | ObjectABC | IndicatorModel):
        """Stage indicator on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = IndicatorModel(**data)

        if not isinstance(data, IndicatorModel):
            ex_msg = 'Invalid type passed in to stage_associated_indicator'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.associated_indicators.data.append(data)  # type: ignore

    def stage_attribute(self, data: dict | ObjectABC | CaseAttributeModel):
        """Stage attribute on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = CaseAttributeModel(**data)

        if not isinstance(data, CaseAttributeModel):
            ex_msg = 'Invalid type passed in to stage_attribute'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.attributes.data.append(data)  # type: ignore

    def stage_note(self, data: dict | ObjectABC | NoteModel):
        """Stage note on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = NoteModel(**data)

        if not isinstance(data, NoteModel):
            ex_msg = 'Invalid type passed in to stage_note'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.notes.data.append(data)  # type: ignore

    def stage_tag(self, data: dict | ObjectABC | TagModel):
        """Stage tag on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = TagModel(**data)

        if not isinstance(data, TagModel):
            ex_msg = 'Invalid type passed in to stage_tag'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.tags.data.append(data)  # type: ignore

    def stage_task(self, data: dict | ObjectABC | TaskModel):
        """Stage task on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = TaskModel(**data)

        if not isinstance(data, TaskModel):
            ex_msg = 'Invalid type passed in to stage_task'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.tasks.data.append(data)  # type: ignore

    def stage_user_access(self, data: dict | ObjectABC | UserModel):
        """Stage user on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = UserModel(**data)

        if not isinstance(data, UserModel):
            ex_msg = 'Invalid type passed in to stage_user_access'
            raise RuntimeError(ex_msg)  # noqa: TRY004
        data._staged = True  # noqa: SLF001
        self.model.user_access.data.append(data)  # type: ignore


class Cases(ObjectCollectionABC):
    """Cases Collection.

    # Example of params input
    {
        "result_limit": 100,  # Limit the retrieved results.
        "result_start": 10,  # Starting count used for pagination.
        "fields": ["caseId", "summary"]  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = CasesModel(**kwargs)
        self.type_ = 'cases'

    def __iter__(self) -> Iterator[Case]:
        """Return CM objects."""
        return self.iterate(base_class=Case)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.CASES.value

    @property
    def filter(self) -> CaseFilter:
        """Return the type specific filter object."""
        return CaseFilter(self.tql)
